#!/usr/bin/env python3

## Classes to explain models predictions
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Classes :
# - Explainer -> Parent class for the explainers
# - LimeExplainer -> Lime Explainer wrapper class

import os
import logging
import tempfile
import numpy as np
import pandas as pd
from PIL import Image
from typing import Type, Union, Any
from lime.lime_image import LimeImageExplainer

from .. import utils
from ..preprocessing import preprocess
from ..models_training.classifiers.model_classifier import ModelClassifierMixin  # type: ignore
# ModelClassifierMixin import must be ignored for mypy as the Mixin class is itself ignored (too complicated to manage)


class Explainer:
    '''Parent class for the explainers'''

    def __init__(self, *args, **kwargs) -> None:
        '''Initialization of the parent class'''
        self.logger = logging.getLogger(__name__)

    def explain_instance(self, content: Image.Image, **kwargs) -> Any:
        '''Explains a prediction

        Args:
            content (Image.Image): Image to be explained
        Returns:
            (?): An explanation object
        '''
        raise NotImplementedError("'explain_instance' needs to be overridden")

    def explain_instance_as_html(self, content: Image.Image, **kwargs) -> str:
        '''Explains a prediction - returns an HTML object

        Args:
            content (Image.Image): Image to be explained
        Returns:
            str: An HTML code with the explanation
        '''
        raise NotImplementedError("'explain_instance_as_html' needs to be overridden")

    def explain_instance_as_json(self, content: Image.Image, **kwargs) -> Union[dict, list]:
        '''Explains a prediction - returns an JSON serializable object

        Args:
            content (str): Text to be explained
        Returns:
            str: A JSON serializable object with the explanation
        '''
        raise NotImplementedError("'explain_instance_as_json' needs to be overridden")

class LimeExplainer(Explainer):
    '''Lime Explainer wrapper class'''

    def __init__(self, model: Type[ModelClassifierMixin], model_conf: dict) -> None:
        ''' Initialization

        Args:
            model: A model instance with predict & predict_proba functions, and list_classes attribute
            model_conf (dict): The model's configuration
        Raises:
            ValueError: If the provided model is not a classifier
            TypeError: If the provided model does not implement a `predict_proba` function
            TypeError: If the provided model does not have a `list_classes` attribute
        '''
        super().__init__()
        pred_proba_op = getattr(model, "predict_proba", None)

        # Check classifier
        if not model.model_type == 'classifier':
            raise ValueError("LimeExplainer only supported with classifier models")
        # Check needed methods
        if pred_proba_op is None or not callable(pred_proba_op):
            raise TypeError("The supplied model must implement a predict_proba() function")
        if getattr(model, "list_classes", None) is None:
            raise TypeError("The supplied model must have a list_classes attribute")

        self.model = model
        self.model_conf = model_conf
        self.class_names = self.model.list_classes
        # Our explainers will explain a prediction for a given class / label
        # These atributes are set on the fly
        self.current_class_index = 0
        # Create the explainer
        self.explainer = LimeImageExplainer()

    def classifier_fn(self, content_arrays: np.ndarray) -> np.ndarray:
        '''Function to get probabilities from a list of (not preprocessed) images

        Args:
            content_arrays (np.ndarray): images to be considered
        Returns:
            np.array: probabilities
        '''
        # Get preprocessor
        if 'preprocess_str' in self.model_conf.keys():
            preprocess_str = self.model_conf['preprocess_str']
        else:
            preprocess_str = "no_preprocess"
        preprocessor = preprocess.get_preprocessor(preprocess_str)
        # Preprocess images
        images = [Image.fromarray(img, 'RGB') for img in content_arrays]
        images_preprocessed = preprocessor(images)
        # Temporary folder
        with tempfile.TemporaryDirectory(dir=utils.get_data_path()) as tmp_folder:
            # Save images
            images_path = [os.path.join(tmp_folder, f'image_{i}.png') for i in range(len(images_preprocessed))]
            for i, img_preprocessed in enumerate(images_preprocessed):
                img_preprocessed.save(images_path[i], format='PNG')
            # Get predictions
            df = pd.DataFrame({'file_path': images_path})
            probas = self.model.predict_proba(df)
        # Return probas
        return probas

    def explain_instance(self, content: Image.Image, class_index: Union[int, None] = None,
                         num_samples: int = 100, batch_size: int = 100, hide_color=0, **kwargs):
        '''Explains a prediction

        This function calls the Lime module. It generates neighborhood data by randomly perturbing features from the instance.
        Then, it learns locally weighted linear models on this neighborhood data to explain each of the classes in an interpretable way.

        Args:
            img (Image.Image): Image to be explained
        Kwargs:
            class_index (int): for classification only. Class or label index to be considered.
            num_samples (int): size of the neighborhood to learn the linear model (cf. Lime documentation)
            batch_size (int): classifier_fn will be called on batches of this size (cf. Lime documentation)
            hide_color (?): TODO
        Returns:
            (?): An explanation object
        '''
        # Set index
        if class_index is not None:
            self.current_class_index = class_index
        else:
            self.current_class_index = 1  # Def to 1
        # Get explanations (images must be convert into rgb, then into np array)
        return self.explainer.explain_instance(np.array(content.convert('RGB')), self.classifier_fn,
                                               labels=(self.current_class_index,),
                                               num_samples=num_samples, batch_size=batch_size,
                                               hide_color=hide_color, top_labels=None)

    def explain_instance_as_html(self, content: Image.Image, **kwargs) -> str:
        '''Explains a prediction - returns an HTML object
        ** NOT IMPLEMENTED **
        '''
        raise NotImplementedError("'explain_instance_as_html' is not defined for LimeExplainer")

    def explain_instance_as_json(self, content: Image.Image, **kwargs) -> Union[dict, list]:
        '''Explains a prediction - returns an JSON serializable object
        ** NOT IMPLEMENTED **
        '''
        raise NotImplementedError("'explain_instance_as_json' is not defined for LimeExplainer")

if __name__ == '__main__':
    logger = logging.getLogger(__name__)
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")



#!/usr/bin/env python3
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""Config global settings

This module handle global app configuration
"""


import pkg_resources  # type: ignore
from pydantic import BaseSettings


class Settings(BaseSettings):
    app_name: str = "{{package_name}}"
    app_version: str = pkg_resources.get_distribution("{{package_name}}").version
    api_entrypoint: str = "/{{package_name}}/rs/v1"
    log_level: str = "INFO"

    class Config:
        env_file = ".env"


settings = Settings()



#!/usr/bin/env python3
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Libs unittest
import unittest
from unittest.mock import Mock
from unittest.mock import patch

# Utils libs
import os
import json
import shutil
import dill as pickle
import numpy as np
import pandas as pd

import tensorflow
import tensorflow.keras as keras
from tensorflow.keras.preprocessing.text import Tokenizer

from {{package_name}} import utils
from {{package_name}}.models_training.models_tensorflow.model_embedding_lstm import ModelEmbeddingLstm

# Disable logging
import logging
logging.disable(logging.CRITICAL)


def remove_dir(path):
    if os.path.isdir(path): shutil.rmtree(path)


class ModelEmbeddingLstmTests(unittest.TestCase):
    '''Main class to test model_embedding_lstm'''

    def setUp(self):
        '''Setup fonction -> we create a mock embedding'''
        # Change directory to script directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)
        # Check if data folder exists
        data_path = utils.get_data_path()
        if not os.path.exists(data_path):
            os.mkdir(data_path)
        # Create a mock embedding
        fake_embedding = {'toto': [0.25, 0.90, 0.12], 'titi': [0.85, 0.12, 0.8], 'test': [0.5, 0.6, 0.1],
                          'pas': [0.82, 0.90, 0.13], 'ici': [0.65, 0.01, 0.01], 'rien': [0.25, 0.02, 0.98],
                          'cela': [0.5, 0.5, 0.5]}
        fake_path = os.path.join(data_path, 'fake_embedding.pkl')
        if os.path.exists(fake_path):
            os.remove(fake_path)
        with open(fake_path, 'wb') as f:
            pickle.dump(fake_embedding, f, pickle.HIGHEST_PROTOCOL)

    def tearDown(self):
        '''Cleaning fonction -> we delete the mock embedding'''
        data_path = utils.get_data_path()
        fake_path = os.path.join(data_path, 'fake_embedding.pkl')
        if os.path.exists(fake_path):
            os.remove(fake_path)

    def check_weights_equality(self, model_1, model_2):
        self.assertEqual(len(model_1.model.weights), len(model_2.model.weights))
        for layer_nb in range(len(model_1.model.weights)):
            self.assertEqual(model_1.model.weights[layer_nb].numpy().shape, model_2.model.weights[layer_nb].numpy().shape)
        for layer_nb, x1, x2 in [(5, 1, 100), (9, 42, 132)]:
            self.assertAlmostEqual(model_1.model.weights[layer_nb].numpy()[x1, x2], model_2.model.weights[layer_nb].numpy()[x1, x2])

    def test01_model_embedding_lstm_init(self):
        '''Test of {{package_name}}.models_training.models_tensorflow.model_embedding_lstm.ModelEmbeddingLstm.__init__'''
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        # Init., test all parameters
        model = ModelEmbeddingLstm(model_dir=model_dir)
        self.assertEqual(model.model_dir, model_dir)
        self.assertTrue(os.path.isdir(model_dir))
        self.assertTrue(model.tokenizer is None)
        # We test display_if_gpu_activated and _is_gpu_activated just by calling them
        self.assertTrue(type(model._is_gpu_activated()) == bool)
        model.display_if_gpu_activated()
        remove_dir(model_dir)

        #
        model = ModelEmbeddingLstm(model_dir=model_dir, max_sequence_length=20)
        self.assertEqual(model.model_dir, model_dir)
        self.assertTrue(os.path.isdir(model_dir))
        self.assertEqual(model.max_sequence_length, 20)
        self.assertTrue(model.tokenizer is None)
        remove_dir(model_dir)

        #
        model = ModelEmbeddingLstm(model_dir=model_dir, max_words=100)
        self.assertEqual(model.model_dir, model_dir)
        self.assertTrue(os.path.isdir(model_dir))
        self.assertEqual(model.max_words, 100)
        self.assertTrue(model.tokenizer is None)
        remove_dir(model_dir)

        #
        model = ModelEmbeddingLstm(model_dir=model_dir, padding='post')
        self.assertEqual(model.model_dir, model_dir)
        self.assertTrue(os.path.isdir(model_dir))
        self.assertEqual(model.padding, 'post')
        self.assertTrue(model.tokenizer is None)
        remove_dir(model_dir)

        #
        model = ModelEmbeddingLstm(model_dir=model_dir, truncating='pre')
        self.assertEqual(model.model_dir, model_dir)
        self.assertTrue(os.path.isdir(model_dir))
        self.assertEqual(model.truncating, 'pre')
        self.assertTrue(model.tokenizer is None)
        remove_dir(model_dir)

        #
        model = ModelEmbeddingLstm(model_dir=model_dir, tokenizer_filters="!;/")
        self.assertEqual(model.model_dir, model_dir)
        self.assertTrue(os.path.isdir(model_dir))
        self.assertEqual(model.tokenizer_filters, "!;/")
        self.assertTrue(model.tokenizer is None)
        # We also test if tokenizer_filters works correctly
        x_train = ['test titi toto', 'toto', 'titi test test toto']
        x_train_prepared = model._prepare_x_train(x_train)
        self.assertFalse(model.tokenizer is None)
        self.assertEqual(model.tokenizer.filters, "!;/")
        remove_dir(model_dir)

        # Check errors
        with self.assertRaises(ValueError):
            model = ModelEmbeddingLstm(model_dir=model_dir, padding='toto')
        with self.assertRaises(ValueError):
            model = ModelEmbeddingLstm(model_dir=model_dir, truncating='toto')

    def test02_model_embedding_lstm_predict_proba(self):
        '''Test of {{package_name}}.models_training.models_tensorflow.model_embedding_lstm.ModelEmbeddingLstm.predict_proba'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        x_train = np.array(["ceci est un test", "pas cela", "cela non plus", "ici test", "là, rien!"] * 100)
        x_test = np.array(["cela est un test", "ni cela", "non plus", "ici test", "là, rien de rien!"] * 100)
        y_train_mono = np.array([0, 1, 0, 1, 2] * 100)
        y_test_mono = y_train_mono.copy()
        y_train_multi = pd.DataFrame({'test1': [0, 0, 0, 1, 0] * 100, 'test2': [1, 0, 0, 0, 0] * 100, 'test3': [0, 0, 0, 1, 0] * 100})
        y_test_multi = y_train_multi.copy()
        cols = ['test1', 'test2', 'test3']

        # Mono-label
        model = ModelEmbeddingLstm(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False,
                                   max_sequence_length=10, max_words=100,
                                   padding='pre', truncating='post',
                                   embedding_name='fake_embedding.pkl')
        model.fit(x_train, y_train_mono)
        preds = model.predict_proba(x_train)
        self.assertEqual(preds.shape, (len(x_train), 3))
        preds = model.predict_proba('test')
        self.assertEqual([elem for elem in preds], [elem for elem in model.predict_proba(['test'])[0]])
        remove_dir(model_dir)

        #
        model = ModelEmbeddingLstm(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False,
                                   max_sequence_length=10, max_words=100,
                                   padding='pre', truncating='post',
                                   embedding_name='fake_embedding.pkl')
        model.fit(x_train, y_train_mono)
        preds = model.predict_proba(x_train, experimental_version=True)
        self.assertEqual(preds.shape, (len(x_train), 3))
        preds = model.predict_proba('test', experimental_version=True)
        self.assertEqual([elem for elem in preds], [elem for elem in model.predict_proba(['test'], experimental_version=True)[0]])
        remove_dir(model_dir)

        # Multi-labels
        model = ModelEmbeddingLstm(model_dir=model_dir, batch_size=8, epochs=2, multi_label=True,
                                   max_sequence_length=10, max_words=100,
                                   padding='pre', truncating='post',
                                   embedding_name='fake_embedding.pkl')
        model.fit(x_train, y_train_multi[cols])
        preds = model.predict_proba(x_train)
        self.assertEqual(preds.shape, (len(x_train), len(cols)))
        preds = model.predict_proba('test')
        self.assertEqual([elem for elem in preds], [elem for elem in model.predict_proba(['test'])[0]])
        remove_dir(model_dir)

        #
        model = ModelEmbeddingLstm(model_dir=model_dir, batch_size=8, epochs=2, multi_label=True,
                                   max_sequence_length=10, max_words=100,
                                   padding='pre', truncating='post',
                                   embedding_name='fake_embedding.pkl')
        model.fit(x_train, y_train_multi[cols])
        preds = model.predict_proba(x_train, experimental_version=True)
        self.assertEqual(preds.shape, (len(x_train), len(cols)))
        preds = model.predict_proba('test', experimental_version=True)
        self.assertEqual([elem for elem in preds], [elem for elem in model.predict_proba(['test'], experimental_version=True)[0]])
        remove_dir(model_dir)

        # Model needs to be fitted
        with self.assertRaises(AttributeError):
            model = ModelEmbeddingLstm(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False,
                                       max_sequence_length=10, max_words=100,
                                       padding='pre', truncating='post',
                                       embedding_name='fake_embedding.pkl')
            model.predict_proba('test')
        remove_dir(model_dir)

    def test03_model_embedding_lstm_prepare_x_train(self):
        '''Test of {{package_name}}.models_training.models_tensorflow.model_embedding_lstm.ModelEmbeddingLstm._prepare_x_train'''

        # Create model
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        max_sequence_length = 10
        model = ModelEmbeddingLstm(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False,
                                   max_sequence_length=max_sequence_length, max_words=100,
                                   padding='pre', truncating='post',
                                   embedding_name='fake_embedding.pkl')

        # Nominal case
        x_train = ['test titi toto', 'toto', 'titi test test toto']
        x_train_prepared = model._prepare_x_train(x_train)
        # We can't easily test the results, too many dependences
        self.assertEqual(x_train_prepared.shape[0], len(x_train))
        self.assertEqual(x_train_prepared.shape[1], max_sequence_length)

        # Clean
        remove_dir(model_dir)

    def test04_model_embedding_lstm_prepare_x_test(self):
        '''Test of {{package_name}}.models_training.models_tensorflow.model_embedding_lstm.ModelEmbeddingLstm._prepare_x_test'''

        # Create model
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        max_sequence_length = 10
        model = ModelEmbeddingLstm(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False,
                                   max_sequence_length=max_sequence_length, max_words=100,
                                   padding='pre', truncating='post',
                                   embedding_name='fake_embedding.pkl')

        # Nominal case
        x_test = ['test titi toto', 'toto', 'titi test test toto']
        model._prepare_x_train(x_test)  # We force the creation of the tokenizer
        x_test_prepared = model._prepare_x_test(x_test)
        # We can't easily test the results, too many dependences
        self.assertEqual(x_test_prepared.shape[0], len(x_test))
        self.assertEqual(x_test_prepared.shape[1], max_sequence_length)

        # Clean
        remove_dir(model_dir)

    def test05_model_embedding_lstm_get_model(self):
        '''Test of {{package_name}}.models_training.models_tensorflow.model_embedding_lstm.ModelEmbeddingLstm._get_model'''

        # Create model
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        model = ModelEmbeddingLstm(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False,
                                   max_sequence_length=10, max_words=100,
                                   padding='pre', truncating='post',
                                   embedding_name='fake_embedding.pkl')

        # Nominal case
        x_train = ['test titi toto', 'toto', 'titi test test toto']
        model._prepare_x_train(x_train)  # We force the creation of the tokenizer
        model.list_classes = ['a', 'b']  # We force the creation of a list of classes
        model_res = model._get_model()
        self.assertTrue(isinstance(model_res, keras.Model))

        # With custom Tokenizer
        tokenizer = Tokenizer()
        tokenizer.fit_on_texts(['toto', 'test', 'tata'])
        model_res = model._get_model(custom_tokenizer=tokenizer)
        self.assertTrue(isinstance(model_res, keras.Model))

        # Clean
        remove_dir(model_dir)

    def test06_model_embedding_lstm_save(self):
        '''Test of the method save of {{package_name}}.models_training.models_tensorflow.model_embedding_lstm.ModelEmbeddingLstm'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        max_sequence_length = 10
        max_words = 100
        padding = 'post'
        truncating = 'pre'
        tokenizer_filters = "!;/"

        # Nominal case - without tokenizer
        model = ModelEmbeddingLstm(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False,
                                   max_sequence_length=max_sequence_length, max_words=100,
                                   padding=padding, truncating=truncating, tokenizer_filters=tokenizer_filters,
                                   embedding_name='fake_embedding.pkl')
        model.save(json_data={'test': 8})
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        # self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'best.hdf5'))) -> no model trained
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'model_embedding_lstm.pkl')))
        self.assertFalse(os.path.exists(os.path.join(model.model_dir, 'embedding_tokenizer.pkl')))
        with open(os.path.join(model.model_dir, 'configurations.json'), 'r', encoding='{{default_encoding}}') as f:
            configs = json.load(f)
        self.assertEqual(configs['test'], 8)
        self.assertEqual(configs['max_sequence_length'], max_sequence_length)
        self.assertEqual(configs['max_words'], max_words)
        self.assertEqual(configs['padding'], padding)
        self.assertEqual(configs['truncating'], truncating)
        self.assertEqual(configs['tokenizer_filters'], tokenizer_filters)
        remove_dir(model_dir)

        # Nominal case - with tokenizer
        tokenizer = Tokenizer()
        tokenizer.fit_on_texts(['toto', 'test', 'tata'])
        model = ModelEmbeddingLstm(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False,
                                   max_sequence_length=max_sequence_length, max_words=100,
                                   padding=padding, truncating=truncating, tokenizer_filters=tokenizer_filters,
                                   embedding_name='fake_embedding.pkl')
        model.tokenizer = tokenizer
        model.save(json_data={'test': 8})
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        # self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'best.hdf5'))) -> no model trained
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'model_embedding_lstm.pkl')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'embedding_tokenizer.pkl')))
        with open(os.path.join(model.model_dir, 'configurations.json'), 'r', encoding='{{default_encoding}}') as f:
            configs = json.load(f)
        self.assertEqual(configs['test'], 8)
        self.assertEqual(configs['max_sequence_length'], max_sequence_length)
        self.assertEqual(configs['max_words'], max_words)
        self.assertEqual(configs['padding'], padding)
        self.assertEqual(configs['truncating'], truncating)
        self.assertEqual(configs['tokenizer_filters'], tokenizer_filters)
        remove_dir(model_dir)

        # Nominal case - with tokenizer, but level_save = 'LOW'
        tokenizer = Tokenizer()
        tokenizer.fit_on_texts(['toto', 'test', 'tata'])
        model = ModelEmbeddingLstm(model_dir=model_dir, batch_size=8, epochs=2, multi_label=False,
                                   max_sequence_length=max_sequence_length, max_words=100,
                                   padding=padding, truncating=truncating, tokenizer_filters=tokenizer_filters,
                                   embedding_name='fake_embedding.pkl', level_save='LOW')
        model.save(json_data={'test': 8})
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        # self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'best.hdf5'))) -> no model trained
        self.assertFalse(os.path.exists(os.path.join(model.model_dir, 'model_embedding_lstm.pkl')))
        self.assertFalse(os.path.exists(os.path.join(model.model_dir, 'embedding_tokenizer.pkl')))
        with open(os.path.join(model.model_dir, 'configurations.json'), 'r', encoding='{{default_encoding}}') as f:
            configs = json.load(f)
        self.assertEqual(configs['test'], 8)
        self.assertEqual(configs['max_sequence_length'], max_sequence_length)
        self.assertEqual(configs['max_words'], max_words)
        self.assertEqual(configs['padding'], padding)
        self.assertEqual(configs['truncating'], truncating)
        self.assertEqual(configs['tokenizer_filters'], tokenizer_filters)
        remove_dir(model_dir)

    def test07_model_keras_init_new_instance_from_configs(self):
        '''Test of the method _init_new_instance_from_configs of {{package_name}}.models_training.models_tensorflow.model_embedding_lstm.ModelEmbeddingLstm'''
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Nominal case
        model = ModelEmbeddingLstm(model_dir=model_dir)
        model.save(json_data={'test': 8})
        configs = model.load_configs(model_dir=model_dir)

        new_model = ModelEmbeddingLstm._init_new_instance_from_configs(configs=configs)
        self.assertTrue(isinstance(new_model, ModelEmbeddingLstm))
        self.assertEqual(new_model.nb_fit, 0)
        self.assertFalse(new_model.trained)
        for attribute in ['x_col', 'y_col', 'list_classes', 'dict_classes', 'multi_label', 'level_save', 'batch_size', 'epochs',
                          'patience', 'embedding_name', 'max_sequence_length', 'max_words', 'padding', 'truncating', 'tokenizer_filters']:
            self.assertEqual(getattr(model, attribute), getattr(new_model, attribute))
        for attribute in ['validation_split']:
            self.assertAlmostEqual(getattr(model, attribute), getattr(new_model, attribute))
        remove_dir(model_dir)
        remove_dir(new_model.model_dir)

        # Check by changing some attributes
        model = ModelEmbeddingLstm(model_dir=model_dir)
        model.nb_fit = 2
        model.trained = True
        model.x_col = 'coucou'
        model.y_col = 'coucou_2'
        model.list_classes = ['class_1', 'class_2', 'class_3']
        model.dict_classes = {0: 'class_1', 1: 'class_2', 2: 'class_3'}
        model.multi_label = True
        model.level_save = 'MEDIUM'
        model.batch_size = 13
        model.epochs = 42
        model.validation_split = 0.3
        model.patience = 15
        model.embedding_name = 'coucou_embedding'
        model.keras_params = {'coucou':1, 'coucou2': 0.3, 'coucou3':'coucou4'}
        model.max_sequence_length = 10
        model.max_words = 232
        model.padding = 'post'
        model.truncating = 'pre'
        model.tokenizer_filters = 'coucou'
        model.save(json_data={'test': 8})
        configs = model.load_configs(model_dir=model_dir)

        new_model = ModelEmbeddingLstm._init_new_instance_from_configs(configs=configs)
        self.assertTrue(isinstance(new_model, ModelEmbeddingLstm))
        self.assertEqual(new_model.nb_fit, 2)
        self.assertTrue(new_model.trained)
        for attribute in ['x_col', 'y_col', 'list_classes', 'dict_classes', 'multi_label', 'level_save', 'batch_size', 'epochs',
                          'patience', 'embedding_name', 'max_sequence_length', 'max_words', 'padding', 'truncating', 'tokenizer_filters']:
            self.assertEqual(getattr(model, attribute), getattr(new_model, attribute))
        for attribute in ['validation_split']:
            self.assertAlmostEqual(getattr(model, attribute), getattr(new_model, attribute))
        self.assertEqual(set(model.keras_params), set(new_model.keras_params))
        self.assertEqual(model.keras_params['coucou'], new_model.keras_params['coucou'])
        self.assertAlmostEqual(model.keras_params['coucou2'], new_model.keras_params['coucou2'])
        self.assertEqual(model.keras_params['coucou3'], new_model.keras_params['coucou3'])
        remove_dir(model_dir)
        remove_dir(new_model.model_dir)

    def test08_model_keras_load_standalone_files(self):
        '''Test of the method _load_standalone_files of {{package_name}}.models_training.models_tensorflow.model_embedding_lstm.ModelEmbeddingLstm'''
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        new_model_dir = os.path.join(os.getcwd(), 'model_test_987654321')
        remove_dir(new_model_dir)

        old_hdf5_path = os.path.join(model_dir, 'best.hdf5')

        # Nominal case with default_model_dir
        model = ModelEmbeddingLstm(model_dir=model_dir, embedding_name='fake_embedding.pkl')
        model.tokenizer = Tokenizer(num_words=model.max_words, filters=model.tokenizer_filters)
        model.list_classes = ['class_1', 'class_2']
        model.model = model._get_model()
        model.model.save(old_hdf5_path)
        model.save(json_data={'test': 8})

        configs = model.load_configs(model_dir=model_dir)
        new_model = ModelEmbeddingLstm._init_new_instance_from_configs(configs=configs)
        new_hdf5_path = os.path.join(new_model.model_dir, 'best.hdf5')
        new_model._load_standalone_files(default_model_dir=model_dir)
        self.assertTrue(os.path.exists(new_hdf5_path))
        self.check_weights_equality(model, new_model)
        self.assertTrue(isinstance(new_model.tokenizer, Tokenizer))

        remove_dir(model_dir)
        remove_dir(new_model.model_dir)

        # Nominal case with explicit paths
        model = ModelEmbeddingLstm(model_dir=model_dir, embedding_name='fake_embedding.pkl')
        model.tokenizer = Tokenizer(num_words=model.max_words, filters=model.tokenizer_filters)
        model.list_classes = ['class_1', 'class_2']
        model.model = model._get_model()
        model.model.save(old_hdf5_path)
        model.save(json_data={'test': 8})

        configs = model.load_configs(model_dir=model_dir)
        new_model = ModelEmbeddingLstm._init_new_instance_from_configs(configs=configs)
        new_hdf5_path = os.path.join(new_model.model_dir, 'best.hdf5')
        new_model._load_standalone_files(tokenizer_path=os.path.join(model_dir, 'embedding_tokenizer.pkl'),
                                         hdf5_path=os.path.join(model_dir, 'best.hdf5'))
        self.assertTrue(os.path.exists(new_hdf5_path))
        self.check_weights_equality(model, new_model)
        self.assertTrue(isinstance(new_model.tokenizer, Tokenizer))

        remove_dir(model_dir)
        remove_dir(new_model.model_dir)

        # Errors
        model = ModelEmbeddingLstm(model_dir=model_dir, embedding_name='fake_embedding.pkl')
        model.tokenizer = Tokenizer(num_words=model.max_words, filters=model.tokenizer_filters)
        model.list_classes = ['class_1', 'class_2']
        model.model = model._get_model()
        model.model.save(old_hdf5_path)
        model.save(json_data={'test': 8})
        os.remove(os.path.join(model_dir, 'embedding_tokenizer.pkl'))

        configs = model.load_configs(model_dir=model_dir)
        new_model = ModelEmbeddingLstm._init_new_instance_from_configs(configs=configs)
        with self.assertRaises(ValueError):
            new_model._load_standalone_files()
        with self.assertRaises(FileNotFoundError):
            new_model._load_standalone_files(default_model_dir=model_dir)

        remove_dir(model_dir)
        remove_dir(new_model.model_dir)


# Perform tests
if __name__ == '__main__':
    # Start tests
    unittest.main()



#!/usr/bin/env python3

## Model Aggregation Classifier

# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Classes :
# - ModelAggregationClassifier -> Model to aggregate several classifier models

import os
import json
import logging
import numpy as np
import pandas as pd
import dill as pickle
from typing import Callable, Union, Tuple, Type

from ... import utils
from .. import utils_models
from ..model_class import ModelClass
from .model_classifier import ModelClassifierMixin  # type: ignore


def proba_argmax(proba: np.ndarray, list_classes: list, **kwargs):
    '''Gives the class corresponding to the argmax of the average of the given probabilities

    Args:
        proba (np.ndarray): The probabilities of each model for each class, array of shape (nb_models, nb_classes)
        list_classes (list): List of classes
    Returns:
        The prediction
    '''
    proba_average = np.sum(proba, axis=0) / proba.shape[0]
    index_class = np.argmax(proba_average)
    return list_classes[index_class]


def majority_vote(predictions: np.ndarray, **kwargs):
    '''Gives the class corresponding to the most present prediction in the given predictions.
    In case of a tie, gives the prediction of the first model involved in the tie
    Args:
        predictions (np.ndarray): The array containing the predictions of each model (shape (n_models))
    Returns:
        The prediction
    '''
    labels, counts = np.unique(predictions, return_counts=True)
    votes = [(label, count) for label, count in zip(labels, counts)]
    votes = sorted(votes, key=lambda x: x[1], reverse=True)
    possible_classes = {vote[0] for vote in votes if vote[1]==votes[0][1]}
    return [prediction for prediction in predictions if prediction in possible_classes][0]


def all_predictions(predictions: np.ndarray, **kwargs) -> np.ndarray:
    '''Calculates the sum of the arrays along axis 0 casts it to bool and then to int.
    Expects a numpy array containing only zeroes and ones.
    When used as an aggregation function, keeps all the prediction of each model (multi-labels)

    Args:
        predictions (np.ndarray) : Array of shape : (n_models, n_classes)
    Return:
        np.ndarray: The prediction
    '''
    return np.sum(predictions, axis=0, dtype=bool).astype(int)


def vote_labels(predictions: np.ndarray, **kwargs) -> np.ndarray:
    '''Gives the result of majority_vote applied on the second axis.
    When used as an aggregation_function, for each class, performs a majority vote for the aggregated models.
    It gives a multi-labels result

    Args:
        predictions (np.ndarray): array of shape : (n_models, n_classes)
    Return:
        np.ndarray: prediction
    '''
    return np.apply_along_axis(majority_vote, 0, predictions)


class ModelAggregationClassifier(ModelClassifierMixin, ModelClass):
    '''Model for aggregating several classifier models'''
    _default_name = 'model_aggregation_classifier'

    _dict_aggregation_function = {'majority_vote': {'aggregation_function': majority_vote, 'using_proba': False, 'multi_label': False},
                                  'proba_argmax': {'aggregation_function': proba_argmax, 'using_proba': True, 'multi_label': False},
                                  'all_predictions': {'aggregation_function': all_predictions, 'using_proba': False, 'multi_label': True},
                                  'vote_labels': {'aggregation_function': vote_labels, 'using_proba': False, 'multi_label': True}}

    def __init__(self, list_models: Union[list, None] = None, aggregation_function: Union[Callable, str] = 'majority_vote',
                 using_proba: bool = False, **kwargs) -> None:
        '''Initialization of the class (see ModelClass for more arguments)
        This model will aggregate the predictions of several model. The user can choose an aggregation function (with **kwargs if not using a list_classes arg)
        from existing ones, or create its own. All models must be either mono label or multi label, we do not accept mixes.
        However, we accept models that do not have the same class / labels. We will consider a meta model with joined classes / labels.

        Kwargs:
            list_models (list) : The list of models to be aggregated (can be None if reloading from standalones)
            aggregation_function (Callable or str) : The aggregation function used (custom function must use **kwargs if not using a list_classes arg)
            using_proba (bool) : Which object is being aggregated (the probabilities or the predictions).
        Raises:
            ValueError: All the aggregated sub_models have not the same multi_label attributes
            ValueError: The multi_label attributes of the aggregated models are inconsistent with multi_label
        '''
        # Init.
        super().__init__(**kwargs)

        # Get logger (must be done after super init)
        self.logger = logging.getLogger(__name__)

        # Set attributes
        self.using_proba = using_proba
        self.aggregation_function = aggregation_function
        
        # Manage submodels
        self.sub_models = list_models  # Transform the list into a list of dictionnaries [{'name': xxx, 'model': xxx}, ...]

        # Check if only classifiers are present
        if False in [isinstance(sub_model['model'], ModelClassifierMixin) for sub_model in self.sub_models]:
            raise ValueError(f"model_aggregation_classifier only accepts classifier models")

        # Check for multi-labels inconsistencies
        set_multi_label = {sub_model['model'].multi_label for sub_model in self.sub_models}
        if len(set_multi_label) > 1:
            raise ValueError(f"All the aggregated sub_models have not the same multi_label attribute")
        if len(set_multi_label.union({self.multi_label})) > 1:
            raise ValueError(f"The multi_label attributes of the aggregated models are inconsistent with self.multi_label = {self.multi_label}.")

        # Set trained & classes info from submodels
        self.trained, self.list_classes, self.dict_classes = self._check_trained()
        # Set nb_fit to 1 if already trained
        if self.trained:
            self.nb_fit = 1

    @property
    def aggregation_function(self):
        '''Getter for aggregation_function'''
        return self._aggregation_function

    @aggregation_function.setter
    def aggregation_function(self, agg_function: Union[Callable, str]):
        '''Setter for aggregation_function
        If a string, try to match a predefined function

        Raises:
            ValueError: If the object aggregation_function is a str but not found in the dictionary of predefined aggregation functions
            ValueError: If the object aggregation_function is incompatible with multi_label
        '''
        # Retrieve aggregation function from dict if a string
        if isinstance(agg_function, str):
            # Get infos
            if agg_function not in self._dict_aggregation_function.keys():
                raise ValueError(f"The aggregation_function ({agg_function}) is not a valid option (must be chosen in {self._dict_aggregation_function.keys()})")
            using_proba = self._dict_aggregation_function[agg_function]['using_proba']
            multi_label = self._dict_aggregation_function[agg_function]['multi_label']
            agg_function = self._dict_aggregation_function[agg_function]['aggregation_function']  # type: ignore
            # Apply checks
            if self.using_proba != using_proba:
                self.logger.warning(f"using_proba {self.using_proba} is incompatible with the selected aggregation function '{agg_function}'. We force using_proba to {using_proba}.")
                self.using_proba = using_proba  # type: ignore
            if self.multi_label != multi_label:
                raise ValueError(f"multi_label {self.multi_label} is incompatible with the selected aggregation function '{agg_function}'.")
        self._aggregation_function = agg_function

    @aggregation_function.deleter
    def aggregation_function(self):
        '''Deleter for aggregation_function'''
        self._aggregation_function = None

    @property
    def sub_models(self):
        '''Getter for sub_models'''
        return self._sub_models

    @sub_models.setter
    def sub_models(self, list_models: Union[list, None] = None):
        '''Setter for sub_models

        Kwargs:
            list_models (list) : The list of models to be aggregated
        '''
        list_models = [] if list_models is None else list_models
        sub_models = []  # Init list of models
        for model in list_models:
            # If a string (a model name), reload it
            if isinstance(model, str):
                real_model, _ = utils_models.load_model(model)
                dict_model = {'name': model, 'model': real_model}
            else:
                dict_model = {'name': os.path.split(model.model_dir)[-1], 'model': model}
            sub_models.append(dict_model.copy())
        self._sub_models = sub_models.copy()

    @sub_models.deleter
    def sub_models(self):
        '''Deleter for sub_models'''
        self._sub_models = None

    def _check_trained(self) -> Tuple[bool, list, dict]:
        '''Checks and sets various attributes related to the fitting of underlying models

        Returns:
            bool: is the aggregation model is considered fitted
            list: list of classes
            dict: dict of classes
        '''
        # Check fitted
        models_trained = {sub_model['model'].trained for sub_model in self.sub_models}
        if len(models_trained) > 0 and all(models_trained):
            # All models trained
            trained = True
            # Set list_classes
            list_classes = list({label for sub_model in self.sub_models for label in sub_model['model'].list_classes})
            list_classes.sort()
            # Set dict_classes based on self.list_classes
            dict_classes = {i: col for i, col in enumerate(list_classes)}
        # No model or not fitted
        else:
            trained, list_classes, dict_classes = False, [], {}
        return trained, list_classes, dict_classes

    def fit(self, x_train, y_train, x_valid=None, y_valid=None, with_shuffle: bool = True, **kwargs) -> None:
        '''Fits the model

        Args:
            x_train (?): Array-like, shape = [n_samples, n_features]
            y_train (?): Array-like, shape = [n_samples, n_targets]
        Kwargs:
            x_valid (?): Array-like, shape = [n_samples, n_features] - not used by sklearn models
            y_valid (?): Array-like, shape = [n_samples, n_targets] - not used by sklearn models
            with_shuffle (bool): If x, y must be shuffled before fitting - not used by sklearn models
        '''
        # Fit each model
        for sub_model in self.sub_models:
            model = sub_model['model']
            if not model.trained:
                model.fit(x_train, y_train, x_valid=x_valid, y_valid=y_valid, with_shuffle=True, **kwargs)

        # Set nb_fit to 1 if not already trained
        if not self.trained:
            self.nb_fit = 1

        # Update attributes
        self.trained, self.list_classes, self.dict_classes = self._check_trained()

    @utils.trained_needed
    def predict(self, x_test, return_proba: bool = False, **kwargs) -> np.ndarray:
        '''Prediction

        Args:
            x_test (?): array-like or sparse matrix of shape = [n_samples, n_features]
            return_proba (bool): If the function should return the probabilities instead of the classes
        Returns:
            np.ndarray: array of shape = [n_samples]
        '''
        # We decide whether to rely on each model's probas or their predictions
        if return_proba:
            return self.predict_proba(x_test)
        else:
            # Get what we want (probas or preds) and use the aggregation function
            if self.using_proba:
                preds_or_probas = self._predict_probas_sub_models(x_test, **kwargs)
            else:
                preds_or_probas = self._predict_sub_models(x_test, **kwargs)
            return np.array([self.aggregation_function(array, list_classes=self.list_classes) for array in preds_or_probas])  # type: ignore

    @utils.trained_needed
    def predict_proba(self, x_test, **kwargs) -> np.ndarray:
        '''Predicts the probabilities on the test set

        Args:
            x_test (?): array-like or sparse matrix of shape = [n_samples, n_features]
        Returns:
            np.ndarray: array of shape = [n_samples, n_classes]
        '''
        probas_sub_models = self._predict_probas_sub_models(x_test, **kwargs)
        # The probas of all models are averaged
        return np.sum(probas_sub_models, axis=1) / probas_sub_models.shape[1]
    
    @utils.trained_needed
    def _predict_sub_models(self, x_test, **kwargs) -> np.ndarray:
        '''Recover the predictions of each model being aggregated

        Args:
            x_test (?): array-like or sparse matrix of shape = [n_samples, n_features]
        Returns:
            np.ndarray: not multi_label : array of shape = [n_samples, nb_model]
                        multi_label : array of shape = [n_samples, nb_model, n_classes]
        '''
        if self.multi_label:
            array_predict = np.array([self._predict_full_list_classes(sub_model['model'], x_test, return_proba=False) for sub_model in self.sub_models])
            array_predict = np.transpose(array_predict, (1, 0, 2))
        else:
            array_predict = np.array([sub_model['model'].predict(x_test) for sub_model in self.sub_models])
            array_predict = np.transpose(array_predict, (1, 0))
        return array_predict

    @utils.trained_needed
    def _predict_probas_sub_models(self, x_test, **kwargs) -> np.ndarray:
        '''Recover the probabilities of each model being aggregated

        Args:
            x_test (?): array-like or sparse matrix of shape = [n_samples, n_features]
        Returns:
            np.ndarray: array of shape = [n_samples, nb_model, nb_classes]
        '''
        array_probas = np.array([self._predict_full_list_classes(sub_model['model'], x_test, return_proba=True) for sub_model in self.sub_models])
        array_probas = np.transpose(array_probas, (1, 0, 2))
        return array_probas

    def _predict_full_list_classes(self, model: Type[ModelClass], x_test, return_proba: bool = False) -> np.ndarray:
        '''For multi_label: adds missing columns in the prediction of model (class missing in their list_classes)
        Or, if return_proba, adds a proba of zero to the missing classes in their list_classes

        Args:
            model (ModelClass): Model to use
            x_test (?): Array-like or sparse matrix of shape = [n_samples, n_features]
            return_proba (bool): If the function should return the probabilities instead of the classes
        Returns:
            np.ndarray: The array with the missing columns added
        '''
        # Get predictions or probas
        preds_or_probas = model.predict(x_test, return_proba=return_proba) # type: ignore

        # Manage each cases. Reorder predictions or probas according to aggregation model list_classes
        # Multi label, proba = True
        # Multi label, proba = False
        # Mono label, proba = True
        if model.multi_label or return_proba:
            df_all = pd.DataFrame(np.zeros((len(preds_or_probas), len(self.list_classes))), columns=self.list_classes)  # type: ignore
            df_model = pd.DataFrame(preds_or_probas, columns=model.list_classes)
            for col in model.list_classes:
                df_all[col] = df_model[col]
            return df_all.to_numpy()
        # Mono label, proba = False
        else:
            return preds_or_probas

    def save(self, json_data: Union[dict, None] = None) -> None:
        '''Saves the model

        Kwargs:
            json_data (dict): Additional configurations to be saved
        '''
        if json_data is None:
            json_data = {}
        # Specific aggregation - save some wanted entries
        train_keys = ['filename', 'filename_valid', 'preprocess_str']
        default_json_data = {key: json_data.get(key, None) for key in train_keys}
        default_json_data['aggregator_dir'] = self.model_dir
        # Save each trained and unsaved model
        for sub_model in self.sub_models:
            path_config = os.path.join(sub_model['model'].model_dir, 'configurations.json')
            if os.path.exists(path_config):
                with open(path_config, 'r', encoding='utf-8') as f:
                    configs = json.load(f)
                    trained = configs.get('trained', False)
                    if not trained:
                        sub_model['model'].save(default_json_data)
            else:
                sub_model['model'].save(default_json_data)

        # Add some specific information
        json_data['list_models_name'] = [sub_model['name'] for sub_model in self.sub_models]
        json_data['using_proba'] = self.using_proba

        # Save aggregation_function if not None & level_save > LOW
        if (self.aggregation_function is not None) and (self.level_save in ['MEDIUM', 'HIGH']):
            # Manage paths
            aggregation_function_path = os.path.join(self.model_dir, "aggregation_function.pkl")
            # Save as pickle
            with open(aggregation_function_path, 'wb') as f:
                pickle.dump(self.aggregation_function, f)

        # Save
        models_list = [sub_model['name'] for sub_model in self.sub_models]
        aggregation_function = self.aggregation_function
        delattr(self, "sub_models")
        delattr(self, "aggregation_function")
        super().save(json_data=json_data)
        setattr(self, "aggregation_function", aggregation_function)
        setattr(self, "sub_models", models_list)  # Setter needs list of models, not sub_models itself

        # Add message in model_upload_instructions.md
        md_path = os.path.join(self.model_dir, f"model_upload_instructions.md")
        line = "/!\\/!\\/!\\/!\\/!\\   The aggregation model is a special model, please ensure that all sub-models and the aggregation model are manually saved together in order to be able to load it .  /!\\/!\\/!\\/!\\/!\\ \n"
        self.prepend_line(md_path, line)

    def prepend_line(self, file_name: str, line: str) -> None:
        ''' Insert given string as a new line at the beginning of a file

        Kwargs:
            file_name (str): Path to file
            line (str): line to insert
        '''
        with open(file_name, 'r+') as f:
            lines = f.readlines()
            lines.insert(0, line)
            f.seek(0)
            f.writelines(lines)

    def reload_from_standalone(self, **kwargs) -> None:
        '''Reloads a model aggregation from its configuration and "standalones" files
           Reloads the sub_models from their files
        Kwargs:
            configuration_path (str): Path to configuration file
            preprocess_pipeline_path (str): Path to preprocess pipeline
            aggregation_function_path (str): Path to aggregation_function_path
        Raises:
            ValueError: If configuration_path is None
            ValueError: If preprocess_pipeline_path is None
            ValueError: If aggregation_function_path is None
            FileNotFoundError: If the object configuration_path is not an existing file
            FileNotFoundError: If the object preprocess_pipeline_path is not an existing file
            FileNotFoundError: If the object aggregation_function_path is not an existing file
        '''
        # Retrieve args
        configuration_path = kwargs.get('configuration_path', None)
        preprocess_pipeline_path = kwargs.get('preprocess_pipeline_path', None)
        aggregation_function_path = kwargs.get('aggregation_function_path', None)

        # Checks
        if configuration_path is None:
            raise ValueError("The argument configuration_path can't be None")
        if preprocess_pipeline_path is None:
            raise ValueError("The argument preprocess_pipeline_path can't be None")
        if aggregation_function_path is None:
            raise ValueError("The argument aggregation_function_path can't be None")
        if not os.path.exists(configuration_path):
            raise FileNotFoundError(f"The file {configuration_path} does not exist")
        if not os.path.exists(preprocess_pipeline_path):
            raise FileNotFoundError(f"The file {preprocess_pipeline_path} does not exist")
        if not os.path.exists(aggregation_function_path):
            raise FileNotFoundError(f"The file {aggregation_function_path} does not exist")

        # Load confs
        with open(configuration_path, 'r', encoding='utf-8') as f:
            configs = json.load(f)
        # Can't set int as keys in json, so need to cast it after reloading
        # dict_classes keys are always ints
        if 'dict_classes' in configs.keys():
            configs['dict_classes'] = {int(k): v for k, v in configs['dict_classes'].items()}
        elif 'list_classes' in configs.keys():
            configs['dict_classes'] = {i: col for i, col in enumerate(configs['list_classes'])}

        # Reload pipeline preprocessing
        with open(preprocess_pipeline_path, 'rb') as f:
            self.preprocess_pipeline = pickle.load(f)

        # Reload aggregation_function_path
        with open(aggregation_function_path, 'rb') as f:
            self.aggregation_function = pickle.load(f)

        # Set class vars
        # self.model_name = # Keep the created name
        # self.model_dir = # Keep the created folder
        self.nb_fit = configs.get('nb_fit', 1)  # Consider one unique fit by default
        self.trained = configs.get('trained', True)  # Consider trained by default
        # Try to read the following attributes from configs and, if absent, keep the current one
        self.sub_models = configs.get('list_models_name', [])  # Transform the list into a list of dictionnaries [{'name': xxx, 'model': xxx}, ...]
        for attribute in ['x_col', 'y_col', 'list_classes', 'dict_classes', 'multi_label', 'level_save',
                          'using_proba']:
            setattr(self, attribute, configs.get(attribute, getattr(self, attribute)))


if __name__ == '__main__':
    logger = logging.getLogger(__name__)
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")


#!/usr/bin/env python3

## Xgboost model
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Classes :
# - ModelXgboostRegressor -> Xgboost model for regression


import os
import json
import logging
import numpy as np
import pandas as pd
import dill as pickle
from typing import Union
from xgboost import XGBRegressor

from sklearn.model_selection import train_test_split

from ... import utils
from ..model_class import ModelClass
from ..regressors.model_regressor import ModelRegressorMixin  # type: ignore


class ModelXgboostRegressor(ModelRegressorMixin, ModelClass):
    '''Xgboost model for regression'''

    _default_name = 'model_xgboost_regressor'

    def __init__(self, xgboost_params: Union[dict, None] = None, early_stopping_rounds: int = 5, validation_split: float = 0.2, **kwargs) -> None:
        '''Initialization of the class  (see ModelClass & ModelRegressorMixin for more arguments)

        Kwargs:
            xgboost_params (dict): Parameters for the Xgboost
                -> https://xgboost.readthedocs.io/en/latest/python/python_api.html#xgboost.XGBRegressor
            early_stopping_rounds (int): Number of rounds for early stopping
            validation_split (float): Validation split fraction.
                Only used if not validation dataset in the fit input
        '''
        # Init.
        super().__init__(**kwargs)

        # Get logger (must be done after super init)
        self.logger = logging.getLogger(__name__)

        # Set parameters
        if xgboost_params is None:
            xgboost_params = {}
        self.xgboost_params = xgboost_params
        self.early_stopping_rounds = early_stopping_rounds
        self.validation_split = validation_split

        # Set objective (if not in params) & init. model
        if 'objective' not in self.xgboost_params.keys():
            self.xgboost_params['objective'] = 'reg:squarederror'
            #  List of objectives https://xgboost.readthedocs.io/en/latest/parameter.html#learning-task-parameters
        self.model = XGBRegressor(**self.xgboost_params)

    def fit(self, x_train, y_train, x_valid=None, y_valid=None, with_shuffle: bool = True, **kwargs) -> None:
        '''Trains the model
           **kwargs permits compatibility with Keras model

        Args:
            x_train (?): Array-like, shape = [n_samples, n_features]
            y_train (?): Array-like, shape = [n_samples,]
        Kwargs:
            x_valid (?): Array-like, shape = [n_samples, n_features]
            y_valid (?): Array-like, shape = [n_samples,]
            with_shuffle (bool): If x, y must be shuffled before fitting
        Raises:
            RuntimeError: If the model is already fitted
        '''
        # TODO: Check if we can continue the training of a xgboost
        if self.trained:
            self.logger.error("We can't train again a xgboost model")
            self.logger.error("Please train a new model")
            raise RuntimeError("We can't train again a xgboost model")

        # We check input format
        x_train, y_train = self._check_input_format(x_train, y_train, fit_function=True)
        # If there is a validation set, we also check the format (but fit_function to False)
        if y_valid is not None and x_valid is not None:
            x_valid, y_valid = self._check_input_format(x_valid, y_valid, fit_function=False)
        # Otherwise, we do a random split
        else:
            self.logger.warning(f"Warning, no validation dataset. We split the training set (fraction valid = {self.validation_split})")
            x_train, x_valid, y_train, y_valid = train_test_split(x_train, y_train, test_size=self.validation_split)

        # Shuffle x, y if wanted
        if with_shuffle:
            p = np.random.permutation(len(x_train))
            x_train = np.array(x_train)[p]
            y_train = np.array(y_train)[p]
        # Else still transform to numpy array
        else:
            x_train = np.array(x_train)
            y_train = np.array(y_train)

        # Also get x_valid & y_valid as numpy
        x_valid = np.array(x_valid)
        y_valid = np.array(y_valid)

        # Set eval set and train
        eval_set = [(x_train, y_train), (x_valid, y_valid)]  # If there’s more than one item in eval_set, the last entry will be used for early stopping.
        prior_objective = self.model.objective
        self.model.fit(x_train, y_train, eval_set=eval_set, early_stopping_rounds=self.early_stopping_rounds, verbose=True)
        post_objective = self.model.objective
        if prior_objective != post_objective:
            self.logger.warning("Warning: the objective function was automatically changed by XGBOOST")
            self.logger.warning(f"Before: {prior_objective}")
            self.logger.warning(f"After: {post_objective}")

        # Set trained
        self.trained = True
        self.nb_fit += 1

    @utils.trained_needed
    def predict(self, x_test: pd.DataFrame, return_proba: bool = False, **kwargs) -> np.ndarray:
        '''Predictions on test set

        Args:
            x_test (pd.DataFrame): DataFrame with the test data to be predicted
        Kwargs:
            return_proba (bool): Present for compatibility with other models. Raises an error if True
        Raises:
            ValueError: If return_proba is True
        Returns:
            (np.ndarray): Array, shape = [n_samples,]
        '''
        # Manage errors
        if return_proba:
            raise ValueError("Models of type model_xgboost_regressor can't handle probabilities")
        # We check input format
        x_test, _ = self._check_input_format(x_test)
        # Warning, "The method returns the model from the last iteration"
        # But : "Predict with X. If the model is trained with early stopping, then best_iteration is used automatically."
        y_pred = self.model.predict(x_test)
        return y_pred

    @utils.trained_needed
    def predict_proba(self, x_test: pd.DataFrame, **kwargs) -> np.ndarray:
        '''Predicts the probabilities on the test set. Here for compatibility

        Args:
            x_test (pd.DataFrame): Array-like, shape = [n_samples]
        Raises:
            ValueError: Model_xgboost_regressor does not implement predict_proba
        Returns:
            (np.ndarray): Array, shape = [n_samples,]
        '''
        # For compatibility
        raise ValueError("Models of type model_xgboost_regressor do not implement the method predict_proba")

    def save(self, json_data: Union[dict, None] = None) -> None:
        '''Saves the model

        Kwargs:
            json_data (dict): Additional configurations to be saved
        '''
        # Save model
        if json_data is None:
            json_data = {}

        json_data['librairie'] = 'xgboost'
        json_data['xgboost_params'] = self.xgboost_params
        json_data['early_stopping_rounds'] = self.early_stopping_rounds
        json_data['validation_split'] = self.validation_split

        # Save xgboost standalone
        if self.level_save in ['MEDIUM', 'HIGH']:
            if self.trained:
                save_path = os.path.join(self.model_dir, 'xbgoost_standalone.model')
                self.model.save_model(save_path)
            else:
                self.logger.warning("Can't save the XGboost in standalone because it hasn't been already fitted")

        # Save
        super().save(json_data=json_data)

    def reload_from_standalone(self, **kwargs) -> None:
        '''Reloads a model from its configuration and "standalones" files
        - /!\\ Experimental /!\\ -

        Kwargs:
            configuration_path (str): Path to configuration file
            xgboost_path (str): Path to standalone xgboost
            preprocess_pipeline_path (str): Path to preprocess pipeline
        Raises:
            ValueError: If configuration_path is None
            ValueError: If xgboost_path is None
            ValueError: If preprocess_pipeline_path is None
            FileNotFoundError: If the object configuration_path is not an existing file
            FileNotFoundError: If the object xgboost_path is not an existing file
            FileNotFoundError: If the object preprocess_pipeline_path is not an existing file
        '''
        # Retrieve args
        configuration_path = kwargs.get('configuration_path', None)
        xgboost_path = kwargs.get('xgboost_path', None)
        preprocess_pipeline_path = kwargs.get('preprocess_pipeline_path', None)

        # Checks
        if configuration_path is None:
            raise ValueError("The argument configuration_path can't be None")
        if xgboost_path is None:
            raise ValueError("The argument xgboost_path can't be None")
        if preprocess_pipeline_path is None:
            raise ValueError("The argument preprocess_pipeline_path can't be None")
        if not os.path.exists(configuration_path):
            raise FileNotFoundError(f"The file {configuration_path} does not exist")
        if not os.path.exists(xgboost_path):
            raise FileNotFoundError(f"The file {xgboost_path} does not exist")
        if not os.path.exists(preprocess_pipeline_path):
            raise FileNotFoundError(f"The file {preprocess_pipeline_path} does not exist")

        # Load confs
        with open(configuration_path, 'r', encoding='{{default_encoding}}') as f:
            configs = json.load(f)
        # Set class vars
        # self.model_name = # Keep the created name
        # self.model_dir = # Keep the created folder
        self.nb_fit = configs.get('nb_fit', 1)  # Consider one unique fit by default
        self.trained = configs.get('trained', True)  # Consider trained by default
        # Try to read the following attributes from configs and, if absent, keep the current one
        for attribute in ['model_type', 'x_col', 'y_col', 'columns_in', 'mandatory_columns',
                          'level_save', 'xgboost_params', 'early_stopping_rounds', 'validation_split']:
            setattr(self, attribute, configs.get(attribute, getattr(self, attribute)))

        # Reload xgboost model
        self.model.load_model(xgboost_path)  # load data

        # Reload pipeline preprocessing
        with open(preprocess_pipeline_path, 'rb') as f:
            self.preprocess_pipeline = pickle.load(f)


if __name__ == '__main__':
    logger = logging.getLogger(__name__)
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")



#!/usr/bin/env python3

## Utils - tools-functions
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Functions :
# - read_folder -> Loads images and classes / bboxes from a directory of images
# - read_folder_object_detection -> Loads images and bboxes from a directory of images - object detection task
# - read_folder_classification -> Loads images and classes from a directory of images - classification task
# - rebuild_metadata_object_detection -> Rebuilds a metadata file from files names and associated bboxes - object detection task
# - rebuild_metadata_classification -> Rebuilds a metadata file from files names and associated classes - classification task
# - display_shape -> Displays the number of lines and columns of a table
# - get_new_column_name -> Gets a new column name from a list of existing ones & a wanted name
# - get_chunk_limits -> Gets chunk limits from a pandas series or dataframe
# - data_agnostic_str_to_list -> Decorator to transform a string into a list of one element.
# - trained_needed -> Decorator to ensure a model has already been trained
# - get_data_path -> Returns the path of the data folder
# - get_models_path -> Returns the path of the models folder
# - get_ressources_path -> Returns the path of the ressources folder
# - get_package_version -> Returns the current package version
# - find_folder_path -> Find a folder in a base folder and its subfolders.
# Classes :
# - HiddenPrints -> Hides all prints
# - DownloadProgressBar -> Displays a progress bar
# - NpEncoder -> JSON encoder to manage numpy objects


import os
import sys
import json
import logging
import numpy as np
import pandas as pd
import pkg_resources
from tqdm import tqdm
from urllib import request
from typing import Tuple, Union, Callable, List, Any

# Get logger
logger = logging.getLogger(__name__)

DIR_PATH = None  # IMPORTANT : THIS VARIABLE MUST BE SET IN PRODUCTION TO POINT TO DATA AND MODELS PARENT FOLDER


def read_folder(folder_path: str, images_ext: tuple = ('.jpg', '.jpeg', '.png'),
                sep: str = '{{default_sep}}', encoding: str = '{{default_encoding}}',
                accept_no_metadata: bool = False) -> Tuple[list, list, str, str]:
    '''Loads images and classes / bboxes from a directory of images

    Args:
        folder_path (str): Directory with the images to be loaded - abs path
    Kwargs:
        images_ext (tuple): Accepted images extensions if automatic detection (i.e. no metadata file)
        sep (str): Separator of the metadata file - if exists
        encoding (str): Encoding of the metadata file - if exists
        accept_no_metadata (bool): If we allow no targets metadata (i.e. returns only file paths, useful for predictions)
    Returns:
        list: List of images path
        list: List of classes associated with images if classification task, bboxes if objet detection task
        str: Name of the prerprocessing pipeline used
        str: Task type ('classification' or 'object_detection')
    '''
    metadata_object_detection = os.path.join(folder_path, 'metadata_bboxes.csv')
    # Object detection
    if os.path.exists(metadata_object_detection):
        logger.info("Object detection task - Loading folder ...")
        path_list, bboxes_list, preprocess_str = read_folder_object_detection(folder_path, images_ext=images_ext,
                                                                              sep=sep, encoding=encoding,
                                                                              accept_no_metadata=accept_no_metadata)
        return path_list, bboxes_list, preprocess_str, 'object_detection'
    # Classifier
    else:
        logger.info("Classification task - Loading folder ...")
        path_list, classes_list, preprocess_str = read_folder_classification(folder_path, images_ext=images_ext,
                                                                             sep=sep, encoding=encoding,
                                                                             accept_no_metadata=accept_no_metadata)
        return path_list, classes_list, preprocess_str, 'classification'


def read_folder_object_detection(folder_path: str, images_ext: tuple = ('.jpg', '.jpeg', '.png'),
                                 sep: str = '{{default_sep}}', encoding: str = '{{default_encoding}}',
                                 accept_no_metadata: bool = False) -> Tuple[list, list, str]:
    '''Loads images and bboxes from a directory of images - object detection task

    Solution 1: usage of a metadata file (metadata_bboxes.csv)
    Solution 2: read images from root directory - no targets

    Args:
        folder_path (str): Directory with the images to be loaded - abs path
    Kwargs:
        images_ext (tuple): Accepted images extensions if automatic detection (i.e. no metadata file)
        sep (str): Separator of the metadata file - if exists
        encoding (str): Encoding of the metadata file - if exists
        accept_no_metadata (bool): If we allow no targets metadata (i.e. returns only file paths, useful for predictions)
    Raises:
        FileNotFoundError: If folder path does not exists
        NotADirectoryError: If the provided path is not a directory
        ValueError: If column 'filename' does not exists in the metadata file
        RuntimeError: If no loading solution found
    Returns:
        list: List of images path
        list: List of bboxes associated with images
        str: Name of the prerprocessing pipeline used
    '''
    logger.info(f"Loading folder {folder_path} ...")

    # Check path exists and it's a directory
    if not os.path.exists(folder_path):
        raise FileNotFoundError(f"The path {folder_path} does not exist")
    if not os.path.isdir(folder_path):
        raise NotADirectoryError(f"{folder_path} is not a valid directory")

    # We first check for a preprocessing file
    preprocess_file = os.path.join(folder_path, 'preprocess_pipeline.conf')
    if os.path.exists(preprocess_file):
        logger.info("Found a preprocessing file")
        with open(preprocess_file, 'r', encoding=encoding) as f:
            preprocess_str = f.readline()
    else:
        logger.info("Can't find a preprocessing file, backup on 'no_preprocess'")
        preprocess_str = 'no_preprocess'

    # Solution 1: we try to load the directory by reading a metadata file
    # This file must be named metadata_bboxes.csv and contain a column `filename`
    metadata_file = os.path.join(folder_path, 'metadata_bboxes.csv')
    if os.path.exists(metadata_file):
        logger.info("Found a metadata file")

        # Loading metadata file
        metadata_df = pd.read_csv(metadata_file, sep=sep, encoding=encoding)
        if 'filename' not in metadata_df.columns:
            raise ValueError("The metadata file must contain a column 'filename'")

        # Retrieving information (path & bboxes)
        filenames = list(metadata_df['filename'].unique())
        path_list = [os.path.join(folder_path, f) for f in filenames]
        # Try to read bboxes
        if all([_ in metadata_df.columns for _ in ['class', 'x1', 'x2', 'y1', 'y2']]):
            bboxes_list = []
            for filename in filenames:
                filtered_bboxes = metadata_df[metadata_df.filename == filename]
                tmp_bboxes_list = []
                for i, row in filtered_bboxes.iterrows():
                    tmp_bboxes_list.append({
                        'class': str(row['class']),  # We ensure all classes are strings
                        'x1': row['x1'],
                        'x2': row['x2'],
                        'y1': row['y1'],
                        'y2': row['y2'],
                    })
                bboxes_list.append(tmp_bboxes_list)
        else:
            logger.info("Can't retrieve bboxes")
            bboxes_list = None

    # Solution 2: if accept no metadata, we retrieve all images inside the root directory
    elif accept_no_metadata:
        folder_list = [f for f in os.listdir(folder_path) if os.path.isfile(os.path.join(folder_path, f))]
        folder_list = [f for f in folder_list if f != 'preprocess_pipeline.conf']  # Do not consider preprocessing file
        folder_list = [f for f in folder_list if f.endswith(images_ext)]  # Keep only images
        path_list = [os.path.join(folder_path, f) for f in folder_list]  # Get abs paths
        bboxes_list = None  # No targets

    # No solution found, raise error
    else:
        raise RuntimeError(f"No loading solution found for folder ({folder_path})")

    # Return
    return path_list, bboxes_list, preprocess_str


def read_folder_classification(folder_path: str, images_ext: tuple = ('.jpg', '.jpeg', '.png'),
                               sep: str = '{{default_sep}}', encoding: str = '{{default_encoding}}',
                               accept_no_metadata: bool = False) -> Tuple[list, list, str]:
    '''Loads images and classes from a directory of images - classification task

    Solution 1: usage of a metadata file (metadata.csv)
    Solution 2: all images at root directory, and prefixed with class names (e.g. class_filename.ext)
    Solution 3: all images saved in class subdirectories
    Solution 4: read images from root directory - no targets

    Args:
        folder_path (str): Directory with the images to be loaded - abs path
    Kwargs:
        images_ext (tuple): Accepted images extensions if automatic detection (i.e. no metadata file)
        sep (str): Separator of the metadata file - if exists
        encoding (str): Encoding of the metadata file - if exists
        accept_no_metadata (bool): If we allow no targets metadata (i.e. returns only file paths, useful for predictions)
    Raises:
        FileNotFoundError: If folder path does not exists
        NotADirectoryError: If the provided path is not a directory
        ValueError: If column 'filename' does not exists in the metadata file
        ValueError: If column 'class' does not exists in the metadata file and accept_no_metadata is False
        RuntimeError: If no loading solution found
    Returns:
        list: List of images path
        list: List of classes associated with images
        str: Name of the prerprocessing pipeline used
    '''
    logger.info(f"Loading folder {folder_path} ...")

    # Check path exists and it's a directory
    if not os.path.exists(folder_path):
        raise FileNotFoundError(f"The path {folder_path} does not exist")
    if not os.path.isdir(folder_path):
        raise NotADirectoryError(f"{folder_path} is not a valid directory")

    # We first check for a preprocessing file
    preprocess_file = os.path.join(folder_path, 'preprocess_pipeline.conf')
    if os.path.exists(preprocess_file):
        logger.info("Found a preprocessing file")
        with open(preprocess_file, 'r', encoding=encoding) as f:
            preprocess_str = f.readline()
    else:
        logger.info("Can't find a preprocessing file, backup on 'no_preprocess'")
        preprocess_str = 'no_preprocess'

    # Solution 1: we try to load the directory by reading a metadata file
    # This file must be named metadata.csv and contain a column `filename`
    metadata_file = os.path.join(folder_path, 'metadata.csv')
    if os.path.exists(metadata_file):
        logger.info("Found a metadata file")

        # Loading metadata file
        metadata_df = pd.read_csv(metadata_file, sep=sep, encoding=encoding)
        if 'filename' not in metadata_df.columns:
            raise ValueError("The metadata file must contain a column 'filename'")

        # Retrieving information (path & classes)
        path_list = list(metadata_df['filename'].values)
        path_list = [os.path.join(folder_path, f) for f in path_list]
        if 'class' in metadata_df.columns:
            classes_list = [str(cl) for cl in metadata_df['class'].values]
        elif accept_no_metadata:
            logger.info("Can't retrieve classes (missing 'class' column in metadata file)")
            classes_list = None
        else:
            raise ValueError("The metadata file must contain a column 'class' with argument `accept_no_metadata` at False")

        # Return here
        return path_list, classes_list, preprocess_str

    # Solution 2: we check if all files are inside the root directory and if they are all prefixed (i.e. prefix_filename.ext)
    folder_list = os.listdir(folder_path)
    folder_list = [f for f in folder_list if f != 'preprocess_pipeline.conf']  # Do not consider preprocessing file
    # Check if all files are images
    if all([f.endswith(images_ext) for f in folder_list]):
        logger.info("Try to load images from root directory")
        path_list = [os.path.join(folder_path, f) for f in folder_list]

        # Check prefixes
        if all([len(f.split('_')) > 1 for f in folder_list]) and all([len(f.split('_')[0]) > 0 for f in folder_list]):
            classes_list = [f.split('_')[0] for f in folder_list]
        else:
            logger.info("Can't retrieve classes (files are not prefixed)")
            classes_list = None

        # Return here
        return path_list, classes_list, preprocess_str

    # Solution 3: check if images are saved in class subdirectories
    folders_elements = os.listdir(folder_path)
    folders_elements = [f for f in folders_elements if f != 'preprocess_pipeline.conf']  # Do not consider preprocessing file
    # Check if only subdirectories
    if all([os.path.isdir(os.path.join(folder_path, f)) for f in folders_elements]):
        # Check if each subdirectories contain images
        if all([all([f2.endswith(images_ext) for f2 in os.listdir(os.path.join(folder_path, f))]) for f in folders_elements]):
            logger.info("Try to load images from class subdirectories")

            # Retrieving information (path & classes)
            path_list = []
            classes_list = []
            for folder in folders_elements:
                tmp_path_list = [os.path.join(folder_path, folder, f) for f in os.listdir(os.path.join(folder_path, folder))]
                tmp_classes_list = [folder] * len(tmp_path_list)
                path_list += tmp_path_list
                classes_list += tmp_classes_list
            return path_list, classes_list, preprocess_str

    # Solution 4: if accept no metadata, we retrieve all images inside the root directory
    elif accept_no_metadata:
        folder_list = [f for f in os.listdir(folder_path) if os.path.isfile(os.path.join(folder_path, f))]
        folder_list = [f for f in folder_list if f != 'preprocess_pipeline.conf']  # Do not consider preprocessing file
        folder_list = [f for f in folder_list if f.endswith(images_ext)]  # Keep only images
        path_list = [os.path.join(folder_path, f) for f in folder_list]  # Get abs paths
        return path_list, None, preprocess_str  # No targets

    # No more solution, raise error
    raise RuntimeError(f"No loading solution found for folder ({folder_path})")


def rebuild_metadata_object_detection(filenames_list: list, bboxes_list: list) -> pd.DataFrame:
    '''Rebuilds a metadata file from files names and associated bboxes - object detection task

    Args:
        filenames_list (list): List of files names (actually a path relative to files parent directory)
        bboxes_list (list): List of bboxes
    Raises:
        ValueError: Both list must be of same length
    Returns:
        pd.DataFrame: The new metadata dataframe
    '''
    if len(filenames_list) != len(bboxes_list):
        raise ValueError("Both list 'filenames_list' & 'bboxes_list' must be of same length") 

    rows = []
    for filename, bboxes in zip(filenames_list, bboxes_list):
        for bbox in bboxes:
            new_row = {'filename': filename, 'class': bbox['class'], 'x1': bbox['x1'], 'x2': bbox['x2'], 'y1': bbox['y1'], 'y2': bbox['y2']}
            rows.append(new_row)

    return pd.DataFrame(rows)


def rebuild_metadata_classification(filenames_list: list, classes_list: list) -> pd.DataFrame:
    '''Rebuilds a metadata file from files names and associated classes - classification task

    Args:
        filenames_list (list): List of files names (actually a path relative to files parent directory)
        classes_list (list): List of classes
    Raises:
        ValueError: Both list must be of same length
    Returns:
        pd.DataFrame: The new metadata dataframe
    '''
    if len(filenames_list) != len(classes_list):
        raise ValueError("Both list 'filenames_list' & 'classes_list' must be of same length")
    return pd.DataFrame({'filename': filenames_list, 'class': classes_list})


def display_shape(df: pd.DataFrame) -> None:
    '''Displays the number of line and of column of a table.

    Args:
        df (pd.DataFrame): Table to parse
    '''
    # Display
    logger.info(f"Number of lines : {df.shape[0]}. Number of columns : {df.shape[1]}.")


def get_chunk_limits(x: Union[pd.DataFrame, pd.Series], chunksize: int = 10000) -> List[Tuple[int]]:
    '''Gets chunk limits from a pandas series or dataframe.

    Args:
        x (pd.Series or pd.DataFrame): Documents to consider
    Kwargs:
        chunksize (int): The chunk size
    Raises:
        ValueError: If the chunk size is negative
    Returns:
        list<tuple>: the chunk limits
    '''
    if chunksize < 0:
        raise ValueError('The object chunksize must not be negative.')
    # Processs
    if chunksize == 0 or chunksize >= x.shape[0]:
        chunks_limits = [(0, x.shape[0])]
    else:
        chunks_limits = [(i * chunksize, min((i + 1) * chunksize, x.shape[0]))
                         for i in range(1 + ((x.shape[0] - 1) // chunksize))]
    return chunks_limits  # type: ignore


def data_agnostic_str_to_list(function: Callable) -> Callable:
    '''Decorator to transform a string into a list of one element.
    DO NOT CAST BACK TO ONE ELEMENT.

    Args:
        function (func): Function to decorate
    Returns:
        function: The decorated function
    '''
    # Get wrapper
    def wrapper(x, *args, **kwargs):
        '''Wrapper'''
        if type(x) == str:
            # Cast str into a single element list
            my_list = [x]
            # Call function
            results = function(my_list, *args, **kwargs)
        else:
            results = function(x, *args, **kwargs)
        # Return
        return results
    return wrapper


def is_ndarray_convertable(obj: Any) -> bool:
    '''Returns True if the object is covertable to a builtin type in the same way a np.ndarray is
    
    Args:
        obj (Any): an object to test
    Returns:
        bool: True if the object is covertable to a list as a np.ndarray is
    '''
    return hasattr(obj, "dtype") and hasattr(obj, "astype") and hasattr(obj, "tolist")


def ndarray_to_builtin_object(obj: Any) -> Any:
    '''Transform a numpy.ndarray like object to a builtin type like int, float or list

    Args:
        obj (Any): An object
    Raises:
        ValueError: Raise a ValueError when obj is not ndarray convertable
    Returns:
        Any: The object converted to a builtin type like int, float or list
    '''
    if is_ndarray_convertable(obj):
        if np.issubdtype(obj.dtype, np.integer):
            return obj.astype(int).tolist()
        elif np.issubdtype(obj.dtype, np.number):
            return obj.astype(float).tolist()
        else:
            return obj.tolist()
    else:
        raise ValueError(f"{obj} is not ndarray convertable")


@data_agnostic_str_to_list
def download_url(urls: list, output_path: str) -> None:
    '''Downloads an object from a list of URLs.
    This function will try every URL until it find an available one.

    Args:
        urls (list): List of URL to try
        output_path (str): Where to save the downloaded object
    Raises:
        ConnectionError: If no URL is available
    '''
    # Start by creating output directory if does not exists
    output_dir = os.path.dirname(output_path)
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    # Test each url
    is_downloaded = False
    for url in urls:
        if not is_downloaded:
            try:
                # From https://stackoverflow.com/questions/15644964/python-progress-bar-and-downloads
                with DownloadProgressBar(unit='B', unit_scale=True, miniters=1, desc=url.split('/')[-1]) as t:
                    request.urlretrieve(url, filename=output_path, reporthook=t.update_to)
                is_downloaded = True  # Download ok
            except Exception:
                logger.warning(f"Can't download from URL {url}.")
    if not is_downloaded:
        raise ConnectionError("Couldn't find a working URL")


def trained_needed(function: Callable) -> Callable:
    '''Decorator to ensure that a model has been trained.

    Args:
        function (func): Function to decorate
    Returns:
        function: The decorated function
    '''
    # Get wrapper
    def wrapper(self, *args, **kwargs):
        '''Wrapper'''
        if not self.trained:
            raise AttributeError(f"The function {function.__name__} can't be called as long as the model hasn't been fitted")
        else:
            return function(self, *args, **kwargs)
    return wrapper


def get_data_path() -> str:
    '''Returns the path to the data folder

    Returns:
        str: Path of the data folder
    '''
    if DIR_PATH is None:
        dir_path = os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))), '{{package_name}}-data')
    else:
        dir_path = os.path.join(os.path.abspath(DIR_PATH), '{{package_name}}-data')
    if not os.path.isdir(dir_path):
        os.mkdir(dir_path)
    return os.path.abspath(dir_path)


def get_models_path() -> str:
    '''Returns the path to the models folder

    Returns:
        str: Path of the models folder
    '''
    if DIR_PATH is None:
        dir_path = os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))), '{{package_name}}-models')
    else:
        dir_path = os.path.join(os.path.abspath(DIR_PATH), '{{package_name}}-models')
    if not os.path.isdir(dir_path):
        os.mkdir(dir_path)
    return os.path.abspath(dir_path)


def get_ressources_path() -> str:
    '''Returns the path to the ressources folder

    Returns:
        str: Path of the ressources folder
    '''
    dir_path = os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))), '{{package_name}}-ressources')
    if not os.path.isdir(dir_path):
        os.mkdir(dir_path)
    return os.path.abspath(dir_path)


def get_package_version() -> str:
    '''Returns the current version of the package

    Returns:
        str: version of the package
    '''
    version = pkg_resources.get_distribution('{{package_name}}').version
    return version


def find_folder_path(folder_name: str, base_folder: Union[str, None] = None) -> str:
    '''Find a folder in a base folder and its subfolders.
    If base_folder is None, considers folder_name as a path and check it exists

    i.e., with the following structure :
    - C:/
        - base_folder/
            - folderA/
                - folderB/
            - folderC/
    find_folder_path(folderA, C:/base_folder) == C:/base_folder/folderA
    find_folder_path(folderB, C:/base_folder) == C:/base_folder/folderA/folderB
    find_folder_path(C:/base_folder/folderC, None) == C:/base_folder/folderC
    find_folder_path(folderB, None) raises an error

    Args:
        folder_name (str): name of the folder to find. If base_folder is None, consider a path instead.
    Kwargs:
        base_folder (str): path of the base folder. If None, consider folder_name as a path.
    Raises:
        FileNotFoundError: If we can't find folder_name in base_folder
        FileNotFoundError: If folder_name is not a valid path (case where base_folder is None)
    Returns:
        str: path to the wanted folder
    '''
    if base_folder is not None:
        folder_path = None
        for path, subdirs, files in os.walk(base_folder):
            for name in subdirs:
                if name == folder_name:
                    folder_path = os.path.join(path, name)
        if folder_path is None:
            raise FileNotFoundError(f"Can't find folder {folder_name} inside {base_folder} and its subfolders")
    else:
        folder_path = folder_name
        if not os.path.exists(folder_path):
            raise FileNotFoundError(f"Can't find folder {folder_path} (considered as a path)")
    return folder_path


# Class (contexte manager) to temporary mute all prints
# It helps with some annoying external libraries
class HiddenPrints:
    '''Hides all prints'''
    def __enter__(self) -> None:
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')
    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        sys.stdout.close()
        sys.stdout = self._original_stdout


# Class (contexte manager) to display a progress bar with downloads
# From https://stackoverflow.com/questions/15644964/python-progress-bar-and-downloads
class DownloadProgressBar(tqdm):
    '''Displays a progress bar'''
    def update_to(self, b: int = 1, bsize: int = 1, tsize: Any = None) -> None:
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)


# JSON encoder to manage numpy objects
class NpEncoder(json.JSONEncoder):
    '''JSON encoder to manage numpy objects'''
    def default(self, obj) -> Any:
        if is_ndarray_convertable(obj):
            return ndarray_to_builtin_object(obj)
        elif isinstance(obj, set):
            return list(obj)
        else:
            return super(NpEncoder, self).default(obj)


if __name__ == '__main__':
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")



# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
from setuptools import setup

# Get package directory
package_directory = os.path.dirname(os.path.abspath(__file__))

# Get package version (env variable or version file + +local)
version_path = os.path.join(package_directory, 'version.txt')
with open(version_path, 'r') as version_file:
    version = version_file.read().strip()
version = os.getenv('VERSION') or f"{version}+local"

# Get package description
readme_path = os.path.join(package_directory, 'README.md')
with open(readme_path, 'r') as readme_file:
    long_description = readme_file.read()

# Setup
setup(
    name='gabarit',
    version=version,
    packages=['gabarit', 'gabarit.template_nlp', 'gabarit.template_num', 'gabarit.template_vision', 'gabarit.template_api'],
    license='AGPL-3.0',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Agence Data Services PE Nantes",
    author_email="contactadsaiframeworks.00619@pole-emploi.fr",
    description="Kickstart your AI project as code",
    url="https://github.com/OSS-Pole-Emploi/gabarit",
    platforms=['windows', 'linux'],
    python_requires='>=3.7',
    include_package_data=True,
    install_requires=[
        'Jinja2==3.0.3'
    ],
    entry_points={
        'console_scripts': [
            'generate_nlp_project = gabarit.template_nlp.generate_nlp_project:main',
            'generate_num_project = gabarit.template_num.generate_num_project:main',
            'generate_vision_project = gabarit.template_vision.generate_vision_project:main',
            'generate_api_project = gabarit.template_api.generate_api_project:main',
        ],
    }
)



from IPython import display


def print_answer(answer: str, language: str = "bash"):
    """Print the answer

    Answers are always code
    """
    return display.Code(answer.strip(), language=language)


def answer_exercice_1():
    return print_answer(
        f"""
# do not forget to activate your virtual environment
# source venv_num_template/bin/activate

python {{package_name}}-scripts/utils/0_split_train_valid_test.py -f wine.csv --perc_train 0.6 --perc_valid 0.2 --perc_test 0.2 --seed 42 --overwrite
"""
    )


def answer_exercice_2():
    return print_answer(
        f"""
# do not forget to activate your virtual environment
# source venv_num_template/bin/activate

python {{package_name}}-scripts/utils/0_create_samples.py -f wine.csv -n 10
"""
    )


def answer_exercice_3():
    return print_answer(
        f"""
# do not forget to activate your virtual environment
# source venv_num_template/bin/activate

python {{package_name}}-scripts/utils/0_sweetviz_report.py -s wine_train.csv -c wine_test.csv
"""
    )


def answer_exercice_4():
    return print_answer(
        f"""
# do not forget to activate your virtual environment
# source venv_num_template/bin/activate

python {{package_name}}-scripts/1_preprocess_data.py -f wine.csv --target_cols target
"""
    )


def answer_exercice_5_preprocess_P2():
    return print_answer(
        f"""
from sklearn.preprocessing import MinMaxScaler

def get_pipelines_dict() -> dict:
    '''Gets a dictionary of available preprocessing pipelines

    Returns:
        dict: Dictionary of preprocessing pipelines
    '''
    pipelines_dict = {{ '{{' }}
        'no_preprocess': ColumnTransformer([('identity', FunctionTransformer(lambda x: x), make_column_selector())]),
        'preprocess_P1': preprocess_P1(),
        'preprocess_P2': preprocess_P2()
    {{ '}}' }}
    return pipelines_dict

def preprocess_P2() -> ColumnTransformer:
    '''Gets "default" preprocessing pipeline
    Returns:
        ColumnTransformer: The pipeline
    '''
    numeric_pipeline = make_pipeline(SimpleImputer(strategy='median'), MinMaxScaler())
    transformers = [
        ('num', numeric_pipeline, make_column_selector(dtype_include='number')),
    ]
    pipeline = ColumnTransformer(transformers, sparse_threshold=0, remainder='drop')
    return pipeline
""",
        language="python",
    )


def answer_exercice_5_preprocess_script():
    return print_answer(
        f"""
# do not forget to activate your virtual environment
# source venv_num_template/bin/activate

python {{package_name}}-scripts/1_preprocess_data.py -f wine.csv -p preprocess_P2 --target_cols target
"""
    )


def answer_exercice_6():
    return print_answer(
        f"""
# do not forget to activate your virtual environment
# source venv_num_template/bin/activate

python {{package_name}}-scripts/1_preprocess_data.py -f wine_train.csv -p preprocess_P1 --target_cols target

# Get last pipeline fitted in {{package_name}}-pipelines
pipeline="$(ls {{package_name}}-pipelines | grep preprocess_P1 | tail -n 1)"

python {{package_name}}-scripts/2_apply_existing_pipeline.py -f wine_valid.csv -p "$pipeline" --target_cols target
"""
    )


def answer_exercice_7():
    return print_answer(
        f"""
# do not forget to activate your virtual environment
# source venv_num_template/bin/activate

python {{package_name}}-scripts/3_training_classification.py -f wine_train_preprocess_P1.csv --filename_valid  wine_valid_preprocess_P1.csv -y target
"""
    )


def answer_exercice_8():
    return print_answer(
        f"""
# in {{package_name}}-scripts/3_training_classification.py :

if model is None:
    model = model_lgbm_classifier.ModelLGBMClassifier(
        x_col=x_col,
        y_col=y_col,
        level_save=level_save,
        preprocess_pipeline=preprocess_pipeline,
        lgbm_params={{ '{{' }}
            "num_leaves": 31,
            "max_depth": -1,
            "learning_rate": 0.1,
            "n_estimators": 100,
        {{ '}}' }},
        multi_label=multi_label,
    )
""",
        language="python",
    )


def answer_exercice_9():
    return print_answer(
        f"""
# do not forget to activate your virtual environment
# source venv_num_template/bin/activate

model="$(ls {{package_name}}-models/model_ridge_classifier | grep model_ridge_classifier_ | tail -n 1)"

python {{package_name}}-scripts/4_predict.py -f wine_test.csv -y target -m "$model"
"""
    )


def answer_exercice_10_preprocess_P3():

    return print_answer(
        f"""
def get_pipelines_dict() -> dict:
    '''Gets a dictionary of available preprocessing pipelines

    Returns:
        dict: Dictionary of preprocessing pipelines
    '''
    pipelines_dict = {{ '{{' }}
        'no_preprocess': ColumnTransformer([('identity', FunctionTransformer(lambda x: x), make_column_selector())]),
        'preprocess_P1': preprocess_P1(),
        'preprocess_P2': preprocess_P2(),
        'preprocess_P3': preprocess_P3()
    {{ '}}' }}
    return pipelines_dict


def preprocess_P3() -> ColumnTransformer:
    '''Gets "default" preprocessing pipeline
    Returns:
        ColumnTransformer: The pipeline
    '''
    numeric_pipeline = make_pipeline(SimpleImputer(strategy="median"), MinMaxScaler())
    cat_pipeline = make_pipeline(SimpleImputer(strategy='most_frequent'), OneHotEncoder(handle_unknown='ignore'))
    transformers = [
        ("num", numeric_pipeline, make_column_selector(pattern="^(?!cultivar).*$", dtype_include="number")),
        ("cat", cat_pipeline, make_column_selector(pattern="cultivar")),
    ]
    pipeline = ColumnTransformer(transformers, sparse_threshold=0, remainder="drop")
    return pipeline
""",
        language="python",
    )


def answer_exercice_10_scripts():

    return print_answer(
        f"""
# do not forget to activate your virtual environment
# source venv_num_template/bin/activate

### train / valid / test splits
python {{package_name}}-scripts/utils/0_split_train_valid_test.py -f wine_reg.csv --perc_train 0.6 --perc_valid 0.2 --perc_test 0.2 --overwrite

### preprocessing training data
python {{package_name}}-scripts/1_preprocess_data.py -f wine_reg_train.csv -p preprocess_P3 --target_cols alcohol

### preprocessing validation data
# Get last pipeline fitted in {{package_name}}-pipelines
pipeline="$(ls {{package_name}}-pipelines | grep preprocess_P3 | tail -n 1)"

python {{package_name}}-scripts/2_apply_existing_pipeline.py -f wine_reg_valid.csv -p "$pipeline" --target_cols alcohol

### train a regressor
python {{package_name}}-scripts/3_training_regression.py -f wine_reg_train_preprocess_P3.csv --filename_valid  wine_reg_valid_preprocess_P3.csv -y alcohol

### predict test data
model="$(ls {{package_name}}-models/model_knn_regressor | grep model_knn_regressor_ | tail -n 1)"

python {{package_name}}-scripts/4_predict.py -f wine_reg_test.csv -y alcohol -m "$model"
"""
    )



#!/usr/bin/env python3

## Preprocessing functions
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import tqdm
import logging
import functools
from PIL import Image
from io import BytesIO
from typing import Callable

from . import manage_white_borders


# Get logger
logger = logging.getLogger(__name__)

# TODO: manage .pdf files ?

# The following preprocessing are "simple" ones and they return "readable" images
# e.g. resizing, cast RGB, orientation ...
# For more specific preprocessing (scaling, brightness shift, zoom, ...), this is usually done
# inside the model (with object like ImageDataGenerator)


def get_preprocessors_dict() -> dict:
    '''Gets a dictionary of available preprocessing

    Returns:
        dict: Dictionary of preprocessing
    '''
    preprocessors_dict = {
        'no_preprocess': lambda x: x,  # - /!\ DO NOT DELETE -> necessary for compatibility /!\ -
        'preprocess_convert_rgb': preprocess_convert_rgb,  # Simple RGB converter
        'preprocess_docs': preprocess_docs,  # Example pipeline with documents (remove white borders, 3/4 ratio, resize 224x224)
    }
    return preprocessors_dict


def get_preprocessor(preprocess_str: str) -> Callable:
    '''Gets a preprocessing (function) from its name

    Args:
        preprocess_str (str): Name of the preprocess
    Raises:
        ValueError: If the name of the preprocess is not known
    Returns:
        Callable: Function to be used for the preprocessing
    '''
    # Process
    preprocessors_dict = get_preprocessors_dict()
    if preprocess_str not in preprocessors_dict.keys():
        raise ValueError(f"The preprocess {preprocess_str} is not known.")
    # Get preprocessor
    preprocessor = preprocessors_dict[preprocess_str]
    # Return
    return preprocessor


def apply_pipeline(images: list, pipeline: list) -> list:
    '''Applies a pipeline (i.e. list of transformations)

    Args:
        images (list): List of images to be transformed
        pipeline (list): List of transformation to be applied
    Returns:
        list: Preprocessed images
    '''
    # Process
    results = None
    for transformer in pipeline:
        if results is None:
            results = transformer(images)
        else:
            results = transformer(results)
    return results


def preprocess_convert_rgb(images: list) -> list:
    '''Applies a simple RGB conversion

    Args:
        images (list): List of images to be transformed
    Returns:
        list: Preprocessed images
    '''
    # Process
    pipeline = [convert_rgb]
    return apply_pipeline(images, pipeline=pipeline)


def preprocess_docs(images: list) -> list:
    '''Applies a list of usual transformations with scanned documents

    Args:
        images (list): List of images to be transformed
    Returns:
        list: Preprocessed images
    '''
    # Process
    pipeline = [convert_rgb, functools.partial(manage_white_borders.remove_white_borders, image_ratio_strategy='fill', image_ratio=0.75, with_rotation=True), resize]
    return apply_pipeline(images, pipeline=pipeline)


def convert_rgb(images: list) -> list:
    '''Converts a list of image into RGB images

    Args:
        images (list): List of images to be converted
    Returns:
        list: RGB images
    '''
    return [im.convert('RGB') for im in images]


def resize(images: list, width: int = 224, height: int = 224) -> list:
    '''Resizes images

    Args:
        images (list): List of images to be resized
    Kwargs:
        width (int): Wanted width
        height (int): Wanted height
    Raises:
        ValueError: If width < 1
        ValueError: If height < 1
    Returns:
        list: Resized images
    '''
    if width < 1:
        raise ValueError('Width must be strictly positive.')
    if height < 1:
        raise ValueError('Height must be strictly positive.')
    # Process images one by one
    results = []
    for i, im in enumerate(tqdm.tqdm(images)):
        results.append(im.resize((width, height)))
    return results


def jpeg_compression(images: list, quality: int = 75) -> list:
    '''Simulates a JPEG compression
    Might be useful for prediction if a model is trained with JPEG compressed images.

    Args:
        images (list): List of images to be compressed
    Kwargs:
        quality (int): Wanted quality
    Returns:
        list: Compressed images
    '''
    # Process images one by one
    results = []
    for i, im in enumerate(tqdm.tqdm(images)):
        out = BytesIO()
        im.save(out, format='JPEG', quality=quality)
        out.seek(0)
        tmp_im = Image.open(out)
        tmp_im.load()  # Cumpulsory to avoid "Too many open files" issue
        results.append(tmp_im)
    return results


if __name__ == '__main__':
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")



#!/usr/bin/env python3
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""This module contains the base Model class

Model is the base model class. It contains a loading and downloading methods that are
used by default to download your model into your Docker container and load it into your
application.

To use a custom model class in your application, create a new module such as
model_awesome.py in this package and write a custom class that overwrite _load_model,
download_model or predict depending on your needs.
"""


import dill as pickle
import logging
from pathlib import Path
from pydantic import BaseSettings
from typing import Any, Tuple, Union


# Manage paths
CURRENT_DIR = Path()
DEFAULT_MODELS_DIR = CURRENT_DIR / "{{package_name}}-models"
DEFAULT_MODEL_PATH = DEFAULT_MODELS_DIR / "model.pkl"

logger = logging.getLogger(__name__)


class ModelSettings(BaseSettings):
    """Download settings

    This class is used for settings management purpose, have a look at the pydantic
    documentation for more details : https://pydantic-docs.helpmanual.io/usage/settings/

    By default, it looks for environment variables (case insensitive) to set the settings
    if a variable is not found, it looks for a file name .env in your working directory
    where you can declare the values of the variables and finally it sets the values
    to the default ones you can see above.
    """

    model_path: Path = DEFAULT_MODEL_PATH

    class Config:
        env_file = ".env"


class Model:
    """Parent model class.

    This class is given as an exemple, you should probably adapt it to your project.
    This class loads the model from a .pkl file. The model must have a predict function.
    """

    def __init__(self):
        '''Init. model class'''
        self._model = None
        self._model_conf = None
        self._model_explainer = None
        self._loaded = False

    def is_model_loaded(self):
        """return the state of the model"""
        return self._loaded

    def loading(self, **kwargs):
        """load the model"""
        self._load_model(**kwargs)
        self._loaded = True

    def predict(self, *args, **kwargs):
        """Make a prediction thanks to the model"""
        return self._model.predict(*args, **kwargs)

    def explain_as_json(self, *args, **kwargs) -> Union[dict, list]:
        """Compute explanations about a prediction and return a JSON serializable object"""
        return self._model_explainer.explain_instance_as_json(*args, **kwargs)

    def explain_as_html(self, *args, **kwargs) -> str:
        """Compute explanations about a prediction and return an HTML report"""
        return self._model_explainer.explain_instance_as_html(*args, **kwargs)

    def _load_model(self, **kwargs) -> None:
        """Load a model from a file

        Returns:
            Tuple[Any, dict]: A tuple containing the model and a dict of metadata about it.
        """
        settings = ModelSettings(**kwargs)

        logger.info(f"Loading the model from {settings.model_path}")
        with settings.model_path.open("rb") as f:
            self._model = pickle.load(f)

        self._model_conf = {
            "model_path": settings.model_path.name,
            "model_name": settings.model_path.stem,
        }
        logger.info(f"Model loaded")

    @staticmethod
    def download_model(**kwargs) -> bool:
        """You should implement a download method to automatically download your model"""
        logger.info("The function download_model is empty. Implement it to automatically download your model.")
        return True



#!/usr/bin/env python3

## Generic model for HuggingFace Transformers
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Classes :
# - ModelHuggingFace -> Generic model for HuggingFace Transformers


import os
import shutil
import logging
import numpy as np
import pandas as pd
import dill as pickle
import seaborn as sns
from copy import deepcopy
import matplotlib.pyplot as plt
from typing import no_type_check, Union, Tuple, Any, Dict
from sklearn.metrics import accuracy_score, f1_score, precision_score, recall_score

import torch
from datasets import Dataset, load_metric
from datasets.arrow_dataset import Batch
from transformers.tokenization_utils_base import BatchEncoding, VERY_LARGE_INTEGER
from transformers import (AutoModelForSequenceClassification, TrainingArguments, Trainer, DataCollatorWithPadding,
                          AutoTokenizer, TextClassificationPipeline, PreTrainedTokenizer, EvalPrediction,
                          TrainerCallback, EarlyStoppingCallback)

from .. import utils
from . import hf_metrics
from .model_class import ModelClass

sns.set(style="darkgrid")

os.environ['DISABLE_MLFLOW_INTEGRATION'] = 'true'  # Needs to be disable for our custom callback on train metrics

HF_CACHE_DIR = utils.get_transformers_path()


class ModelHuggingFace(ModelClass):
    '''Generic model for Huggingface NN'''

    _default_name = 'model_huggingface'

    # TODO: perhaps it would be smarter to have this class behaving as the abstract class for all the model types
    # implemented on the HF hub and to create model specific subclasses.
    # => might change it as use cases grow

    def __init__(self, batch_size: int = 8, epochs: int = 99, validation_split: float = 0.2, patience: int = 5,
                 transformer_name: str = 'Geotrend/distilbert-base-fr-cased', transformer_params: Union[dict, None] = None,
                 trainer_params: Union[dict, None] = None, model_max_length: int = None, **kwargs) -> None:
        '''Initialization of the class (see ModelClass for more arguments)

        Kwargs:
            batch_size (int): Batch size
            epochs (int): Number of epochs
            validation_split (float): Percentage for the validation set split
                Only used if no input validation set when fitting
            patience (int): Early stopping patience
            transformer_name (str) : The name of the transformer backbone to use
            transformer_params (dict): Parameters used by the Transformer model.
                The purpose of this dictionary is for the user to use it as they wants in the _get_model function
                This parameter was initially added in order to do an hyperparameters search
            trainer_params (dict): A set of parameters to be use by the Trainer. It is recommended to use the default params (leave this empty).
        '''
        # TODO: learning rate should be an attribute !
        # Init.
        super().__init__(**kwargs)

        # Get logger (must be done after super init)
        self.logger = logging.getLogger(__name__)

        # Param. model
        self.batch_size = batch_size
        self.epochs = epochs
        self.validation_split = validation_split
        self.patience = patience
        self.transformer_name = transformer_name
        self.model_max_length = model_max_length

        # transformer_params has no use as of 14/12/2022
        # we still leave it for compatibility with keras models and future usage
        self.transformer_params = transformer_params

        # Trainer params
        if trainer_params is None:
            trainer_params = {
                'output_dir': self.model_dir,
                'learning_rate': 2e-5,
                'per_device_train_batch_size': self.batch_size,
                'per_device_eval_batch_size': self.batch_size,
                'num_train_epochs': self.epochs,
                'weight_decay': 0.0,
                'evaluation_strategy': 'epoch',
                'save_strategy': 'epoch',
                'logging_strategy': 'epoch',
                'save_total_limit': 1,
                'load_best_model_at_end': True
            }
        # TODO: maybe we should keep the default dict & only add/replace keys in provided dict ?
        self.trainer_params = trainer_params

        # Model set on fit or on reload
        self.model: Any = None
        self.pipe: Any = None  # Set on first predict

        # Tokenizer set on fit or on reload
        self.tokenizer: Any = None

    def fit(self, x_train, y_train, x_valid=None, y_valid=None, with_shuffle: bool = True, **kwargs) -> None:
        '''Fits the model

        Args:
            x_train (?): Array-like, shape = [n_samples, n_features]
            y_train (?): Array-like, shape = [n_samples, n_targets]
            x_valid (?): Array-like, shape = [n_samples, n_features]
            y_valid (?): Array-like, shape = [n_samples, n_targets]
        Kwargs:
            with_shuffle (bool): If x, y must be shuffled before fitting
                Experimental: We must verify if it works as intended depending on the formats of x and y
                This should be used if y is not shuffled as the split_validation takes the lines in order.
                Thus, the validation set might get classes which are not in the train set ...
        Raises:
            ValueError: If different classes when comparing an already fitted model and a new dataset
        '''
        ##############################################
        # Manage retrain
        ##############################################

        # If a model has already been fitted, we make a new folder in order not to overwrite the existing one !
        # And we save the old conf
        if self.trained:
            # Get src files to save
            src_files = [os.path.join(self.model_dir, "configurations.json")]
            if self.nb_fit > 1:
                for i in range(1, self.nb_fit):
                    src_files.append(os.path.join(self.model_dir, f"configurations_fit_{i}.json"))
            # Change model dir
            self.model_dir = self._get_new_model_dir()
            # Get dst files
            dst_files = [os.path.join(self.model_dir, f"configurations_fit_{self.nb_fit}.json")]
            if self.nb_fit > 1:
                for i in range(1, self.nb_fit):
                    dst_files.append(os.path.join(self.model_dir, f"configurations_fit_{i}.json"))
            # Copies
            for src, dst in zip(src_files, dst_files):
                try:
                    shutil.copyfile(src, dst)
                except Exception as e:
                    self.logger.error(f"Impossible to copy {src} to {dst}")
                    self.logger.error("We still continue ...")
                    self.logger.error(repr(e))

        ##############################################
        # Prepare x_train, x_valid, y_train & y_valid
        # Also extract list of classes
        ##############################################

        # If not multilabel, transform y_train as dummies (should already be the case for multi-labels)
        if not self.multi_label:
            # If len(array.shape)==2, we flatten the array if the second dimension is useless
            if isinstance(y_train, np.ndarray) and len(y_train.shape) == 2 and y_train.shape[1] == 1:
                y_train = np.ravel(y_train)
            if isinstance(y_valid, np.ndarray) and len(y_valid.shape) == 2 and y_valid.shape[1] == 1:
                y_valid = np.ravel(y_valid)
            # Transformation dummies
            y_train_dummies = pd.get_dummies(y_train)
            y_valid_dummies = pd.get_dummies(y_valid) if y_valid is not None else None
            # Important : get_dummies reorder the columns in alphabetical order
            # Thus, there is no problem if we fit again on a new dataframe with shuffled data
            list_classes = list(y_train_dummies.columns)
            # FIX: valid test might miss some classes, hence we need to add them back to y_valid_dummies
            if y_valid_dummies is not None and y_train_dummies.shape[1] != y_valid_dummies.shape[1]:
                for cl in list_classes:
                    # Add missing columns
                    if cl not in y_valid_dummies.columns:
                        y_valid_dummies[cl] = 0
                y_valid_dummies = y_valid_dummies[list_classes]  # Reorder
        # Else keep it as it is
        else:
            y_train_dummies = y_train
            y_valid_dummies = y_valid
            if hasattr(y_train_dummies, 'columns'):
                list_classes = list(y_train_dummies.columns)
            else:
                self.logger.warning(
                    "Can't read the name of the columns of y_train -> inverse transformation won't be possible"
                )
                # We still create a list of classes in order to be compatible with other functions
                list_classes = [str(_) for _ in range(pd.DataFrame(y_train_dummies).shape[1])]

        # Set dict_classes based on list classes
        dict_classes = {i: col for i, col in enumerate(list_classes)}

        # Validate classes if already trained, else set them
        if self.trained:
            if self.list_classes != list_classes:
                raise ValueError("Error: the new dataset does not match with the already fitted model")
            if self.dict_classes != dict_classes:
                raise ValueError("Error: the new dataset does not match with the already fitted model")
        else:
            self.list_classes = list_classes
            self.dict_classes = dict_classes

        # Shuffle x, y if wanted
        # It is advised as validation_split from keras does not shufle the data
        # Hence we might have classes in the validation data that we never met in the training data
        if with_shuffle:
            p = np.random.permutation(len(x_train))
            x_train = np.array(x_train)[p]
            y_train_dummies = np.array(y_train_dummies)[p]
        # Else still transform to numpy array
        else:
            x_train = np.array(x_train)
            y_train_dummies = np.array(y_train_dummies)

        # Also get y_valid_dummies as numpy
        y_valid_dummies = np.array(y_valid_dummies)

        # If no valid set, split train set according to validation_split
        if y_valid is None:
            self.logger.warning(f"Warning, no validation set. The training set will be splitted (validation fraction = {self.validation_split})")
            p = np.random.permutation(len(x_train))
            cutoff = int(len(p) * self.validation_split)
            x_valid = x_train[p[0:cutoff]]
            x_train = x_train[p[cutoff:]]
            y_valid_dummies = y_train_dummies[p[0:cutoff]]
            y_train_dummies = y_train_dummies[p[cutoff:]]

        ##############################################
        # Get model & prepare datasets
        ##############################################

        # Get model (if already fitted, _get_model returns instance model)
        self.model = self._get_model(num_labels=y_train_dummies.shape[1])

        # Get tokenizer (if already fitted, _get_tokenizer returns instance tokenizer)
        self.tokenizer = self._get_tokenizer()

        # Preprocess datasets
        train_dataset = self._prepare_x_train(x_train, y_train_dummies)
        valid_dataset = self._prepare_x_valid(x_valid, y_valid_dummies)

        ##############################################
        # Fit
        ##############################################

        # Fit
        try:
            # TODO: remove the checkpoints !
            # Prepare trainer
            trainer = Trainer(
                model=self.model,
                args=TrainingArguments(**self.trainer_params),
                train_dataset=train_dataset,
                eval_dataset=valid_dataset,
                tokenizer=self.tokenizer,  # Only use for padding, dataset are already preprocessed. Pby not needed as we define a collator.
                data_collator=DataCollatorWithPadding(tokenizer=self.tokenizer),  # Pad batches
                compute_metrics=self._compute_metrics_mono_label if not self.multi_label else self._compute_metrics_multi_label,
                optimizers=self._get_optimizers(),
            )
            # Add callbacks
            trainer.add_callback(MetricsTrainCallback(trainer))
            trainer.add_callback(EarlyStoppingCallback(early_stopping_patience=self.patience))
            # Fit
            trainer.train()
            # Save model & tokenizer
            hf_model_dir = os.path.join(self.model_dir, 'hf_model')
            hf_tokenizer_dir = os.path.join(self.model_dir, 'hf_tokenizer')
            trainer.model.save_pretrained(save_directory=hf_model_dir)
            self.tokenizer.save_pretrained(save_directory=hf_tokenizer_dir)
            # Remove checkpoint dir if save total limit is set to 1 (no need to keep this as we resave the model)
            if self.trainer_params.get('save_total_limit', None) == 1:
                checkpoint_dirs = [_ for _ in os.listdir(self.model_dir) if _.startswith('checkpoint-')]
                if len(checkpoint_dirs) == 0:
                    self.logger.warning("Can't find a checkpoint dir to be removed.")
                else:
                    for checkpoint_dir in checkpoint_dirs:
                        shutil.rmtree(os.path.join(self.model_dir, checkpoint_dir))
        except (RuntimeError, SystemError, SystemExit, EnvironmentError, KeyboardInterrupt, Exception) as e:
            self.logger.error(repr(e))
            raise RuntimeError("Error during model training")

        # Print accuracy & loss if level_save > 'LOW'
        if self.level_save in ['MEDIUM', 'HIGH']:
            # Plot accuracy
            fit_history = trainer.state.log_history
            self._plot_metrics_and_loss(fit_history)
            # Reload best model ?
            # Default trainer has load_best_model_at_end = True
            # Hence we consider the best model is already reloaded

        # Set trained
        self.trained = True
        self.nb_fit += 1

    @utils.data_agnostic_str_to_list
    @utils.trained_needed
    def predict(self, x_test, return_proba: bool = False, **kwargs) -> np.ndarray:
        '''Predictions on test set

        Args:
            x_test (?): Array-like or sparse matrix, shape = [n_samples]
        Kwargs:
            return_proba (bool): If the function should return the probabilities instead of the classes
        Returns:
            (np.ndarray): Array, shape = [n_samples, n_classes]
        '''
        # Predict probas
        predicted_proba = self.predict_proba(x_test)

        # We return the probabilities if wanted
        if return_proba:
            return predicted_proba

        # Finally, we get the classes predictions
        return self.get_classes_from_proba(predicted_proba)

    @utils.data_agnostic_str_to_list
    @utils.trained_needed
    def predict_proba(self, x_test, **kwargs) -> np.ndarray:
        '''Predicts probabilities on the test dataset

        Args:
            x_test (?): Array-like or sparse matrix, shape = [n_samples, n_features]
        Returns:
            (np.ndarray): Array, shape = [n_samples, n_classes]
        '''
        # Does not work with np array nor pandas Series
        if type(x_test) in [np.ndarray, pd.Series]:
            x_test = x_test.tolist()
        # Prepare predict
        if self.model.training:
            self.model.eval()
        if self.pipe is None:
            # Set model on gpu if available
            self.model = self.model.to('cuda') if self._is_gpu_activated() else self.model.to('cpu')
            device = 0 if self._is_gpu_activated() else -1
            self.pipe = TextClassificationPipeline(model=self.model, tokenizer=self.tokenizer, return_all_scores=True, device=device)
        # Predict
        # As we are using the pipeline, we do not need to prepare x_test (done inside the pipeline)
        # However, we still need to set the tokenizer params (truncate & padding)
        tokenizer_kwargs = {'padding': False, 'truncation': True}
        results = np.array(self.pipe(x_test, **tokenizer_kwargs))
        predicted_proba = np.array([[x['score'] for x in x] for x in results])
        return predicted_proba

    def _prepare_x_train(self, x_train, y_train_dummies) -> Dataset:
        '''Prepares the input data for the model - train

        Args:
            x_train (?): Array-like, shape = [n_samples, n_features]
        Returns:
            (datasets.Dataset): Prepared dataset
        '''
        # TMP FIX : https://github.com/OSS-Pole-Emploi/gabarit/issues/98
        # We can't call this function if the tokenizer is not set. We will pby change this object to a property.
        # This isn't really a problem as this function should not be called outside the class & tokenizer is set in the fit function.
        if self.tokenizer is None:
            self.tokenizer = self._get_tokenizer()
        # Check np format (should be the case if using fit)
        if not type(x_train) == np.ndarray:
            x_train = np.array(x_train)
        if not type(y_train_dummies) == np.ndarray:
            y_train_dummies = np.array(y_train_dummies)
        # It seems that HF does not manage dummies targets for non multilabel
        if not self.multi_label:
            labels = np.argmax(y_train_dummies, axis=-1).astype(int).tolist()
        else:
            labels = y_train_dummies.astype(np.float32).tolist()
        return Dataset.from_dict({'text': x_train.tolist(), 'label': labels}).map(self._tokenize_function, batched=True)

    def _prepare_x_valid(self, x_valid, y_valid_dummies) -> Dataset:
        '''Prepares the input data for the model - valid

        Args:
            x_valid (?): Array-like, shape = [n_samples, n_features]
        Returns:
            (datasets.Dataset): Prepared dataset
        '''
        # Same as train (we don't fit any tokenizer)
        return self._prepare_x_train(x_valid, y_valid_dummies)

    def _prepare_x_test(self, x_test) -> Dataset:
        '''Prepares the input data for the model - test

        Args:
            x_test (?): Array-like, shape = [n_samples, n_features]
        Returns:
            (datasets.Dataset): Prepared dataset
        '''
        # Check np format
        if not type(x_test) == np.ndarray:
            x_test = np.array(x_test)
        # /!\ We don't use it as we are using a TextClassificationPipeline
        # yet we are leaving this here in case we need it later
        return Dataset.from_dict({'text': x_test.tolist()}).map(self._tokenize_function, batched=True)

    def _tokenize_function(self, examples: Batch) -> BatchEncoding:
        '''Tokenizes input data

        Args:
            examples (Batch): input data (Dataset Batch)
        Returns:
            BatchEncoding: tokenized data
        '''
        # Padding to False as we will use a Trainer and a DataCollatorWithPadding that will manage padding for us (better limit the memory impact)
        # We leave max_length to None -> backup on model max length
        # https://stackoverflow.com/questions/74657367/how-do-i-know-which-parameters-to-use-with-a-pretrained-tokenizer
        return self.tokenizer(examples["text"], padding=False, truncation=True)

    def _get_model(self, model_path: str = None, num_labels: int = None) -> Any:
        '''Gets a model structure - returns the instance model instead if already defined

        Returns:
            (Any): a HF model
        '''
        # Return model if already set
        if self.model is not None:
            return self.model

        model = AutoModelForSequenceClassification.from_pretrained(
                self.transformer_name if model_path is None else model_path,
                num_labels=len(self.list_classes) if num_labels is None else num_labels,
                problem_type="multi_label_classification" if self.multi_label else "single_label_classification",
                {% if huggingface_proxies is not none %}proxies={{huggingface_proxies}},
                {% endif %}cache_dir=HF_CACHE_DIR)
        # Set model on gpu if available
        model = model.to('cuda') if self._is_gpu_activated() else model.to('cpu')
        return model

    def _get_tokenizer(self, model_path: str = None) -> PreTrainedTokenizer:
        '''Gets a tokenizer

        Returns:
            (PreTrainedTokenizer): a HF tokenizer
        '''
        # Return tokenizer if already set
        if self.tokenizer is not None:
            return self.tokenizer

        tokenizer = AutoTokenizer.from_pretrained(self.transformer_name if model_path is None else model_path,
                                                  {% if huggingface_proxies is not none %}proxies={{huggingface_proxies}},
                                                  {% endif %}cache_dir=HF_CACHE_DIR)

        if self.model_max_length:
            tokenizer.model_max_length = self.model_max_length

        # If the model name is not in tokenizer.max_model_input_sizes it is likely that the attribute model_max_length is not well
        # initialized. If it is set to VERY_LARGE_INTEGER we warn the user that there is a risk of errors with long sequences
        elif self.transformer_name not in tokenizer.max_model_input_sizes and tokenizer.model_max_length == VERY_LARGE_INTEGER:
            self.logger.warning(f"The model name '{self.transformer_name}' is not present in tokenizer.max_model_input_sizes : '{tokenizer.max_model_input_sizes}' "
                                f"and tokenizer.model_max_length is set to VERY_LARGE_INTEGER. You may encounter errors with long sequences. "
                                f"see. https://huggingface.co/transformers/v4.0.1/main_classes/tokenizer.html?highlight=very_large_integer#transformers.PreTrainedTokenizer")

        return tokenizer

    def _get_optimizers(self) -> Tuple[Any, Any]:
        '''Fonction to define the Trainer optimizers
           -> per default return (None, None), i.e. default optimizers (cf HF Trainer doc)

        Returns:
            Tuple (Optimizer, LambdaLR): An optimizer/scheduler couple
        '''
        # e.g.
        # Here, your custom Optimizer / scheduler couple
        # (check https://huggingface.co/docs/transformers/v4.24.0/en/main_classes/optimizer_schedules)
        return (None, None)

    def _compute_metrics_mono_label(self, eval_pred: EvalPrediction) -> dict:
        '''Computes some metrics for mono label cases

        Args:
            eval_pred: predicted & ground truth values to be considered
        Returns:
            dict: dictionnary with computed metrics
        '''
        # Load metrics
        metric_accuracy = load_metric(hf_metrics.accuracy.__file__)
        metric_precision = load_metric(hf_metrics.precision.__file__)
        metric_recall = load_metric(hf_metrics.recall.__file__)
        metric_f1 = load_metric(hf_metrics.f1.__file__)
        # Get predictions
        logits, labels = eval_pred
        predictions = np.argmax(logits, axis=-1)

        # Compute metrics
        accuracy = metric_accuracy.compute(predictions=predictions, references=labels)["accuracy"]
        precision = metric_precision.compute(predictions=predictions, references=labels, average='weighted')["precision"]
        recall = metric_recall.compute(predictions=predictions, references=labels, average='weighted')["recall"]
        f1 = metric_f1.compute(predictions=predictions, references=labels, average='weighted')["f1"]
        # Return dict of metrics
        return {'accuracy': accuracy, 'weighted_precision': precision, 'weighted_recall': recall, 'weighted_f1': f1}

    def _compute_metrics_multi_label(self, eval_pred: EvalPrediction) -> dict:
        '''Computes some metrics for mono label cases

        Args:
            eval_pred: predicted & ground truth values to be considered
        Returns:
            dict: dictionnary with computed metrics
        '''
        # Sigmoid activation (multi_label)
        sigmoid = torch.nn.Sigmoid()
        # Get probas
        logits, labels = eval_pred
        probas = sigmoid(torch.Tensor(logits))
        # Get predictions (probas >= 0.5)
        predictions = np.zeros(probas.shape)
        predictions[np.where(probas >= 0.5)] = 1
        # Compute metrics (we can't use HF metrics, it sucks)
        accuracy = accuracy_score(y_true=labels, y_pred=predictions)  # Must be exact match on all labels
        f1 = f1_score(y_true=labels, y_pred=predictions, average='weighted')
        precision = precision_score(y_true=labels, y_pred=predictions, average='weighted')
        recall = recall_score(y_true=labels, y_pred=predictions, average='weighted')
        # return as dictionary
        return {'accuracy': accuracy, 'weighted_precision': precision, 'weighted_recall': recall, 'weighted_f1': f1}

    def _plot_metrics_and_loss(self, fit_history) -> None:
        '''Plots TrainOutput, for legacy and compatibility purpose

        Arguments:
            fit_history (list) : fit history - actually list of logs
        '''
        # Manage dir
        plots_path = os.path.join(self.model_dir, 'plots')
        if not os.path.exists(plots_path):
            os.makedirs(plots_path)

        # Rework fit_history to better match Keras fit history
        fit_history_dict: Dict[str, list] = {}
        for log in fit_history:
            for key, value in log.items():
                if key not in fit_history_dict.keys():
                    fit_history_dict[key] = [value]
                else:
                    fit_history_dict[key] += [value]

        # Get a dictionnary of possible metrics/loss plots
        metrics_dir = {
            'loss': ['Loss', 'loss'],
            'accuracy': ['Accuracy', 'accuracy'],
            'weighted_f1': ['Weighted F1-score', 'weighted_f1_score'],
            'weighted_precision': ['Weighted Precision', 'weighted_precision'],
            'weighted_recall': ['Weighted Recall', 'weighted_recall'],
        }

        # Plot each available metric
        for metric in metrics_dir.keys():
            if any([f'{dataset}_{metric}' in fit_history_dict.keys() for dataset in ['train_metrics', 'eval']]):
                title = metrics_dir[metric][0]
                filename = metrics_dir[metric][1]
                plt.figure(figsize=(10, 8))
                legend = []
                for dataset in ['train_metrics', 'eval']:
                    if f'{dataset}_{metric}' in fit_history_dict.keys():
                        plt.plot(fit_history_dict[f'{dataset}_{metric}'])
                        legend += ['Train'] if dataset == 'train_metrics' else ['Validation']
                plt.title(f"Model {title}")
                plt.ylabel(title)
                plt.xlabel('Epoch')
                plt.legend(legend, loc='upper left')
                # Save
                filename == f"{filename}.jpeg"
                plt.savefig(os.path.join(plots_path, filename))

                # Close figures
                plt.close('all')

    @no_type_check  # We do not check the type, because it is complicated with managing custom_objects_str
    def save(self, json_data: Union[dict, None] = None) -> None:
        '''Saves the model

        Kwargs:
            json_data (dict): Additional configurations to be saved
        '''
        # Save configuration JSON
        if json_data is None:
            json_data = {}

        json_data['librairie'] = 'huggingface'
        json_data['batch_size'] = self.batch_size
        json_data['epochs'] = self.epochs
        json_data['validation_split'] = self.validation_split
        json_data['patience'] = self.patience
        json_data['transformer_name'] = self.transformer_name
        json_data['transformer_params'] = self.transformer_params
        json_data['trainer_params'] = self.trainer_params
        json_data['model_max_length'] = self.model_max_length

        # Add model structure if not none
        if self.model is not None:
            json_data['hf_model'] = self.model.__repr__()

        if '_get_model' not in json_data.keys():
            json_data['_get_model'] = pickle.source.getsourcelines(self._get_model)[0]
        if '_get_tokenizer' not in json_data.keys():
            json_data['_get_tokenizer'] = pickle.source.getsourcelines(self._get_tokenizer)[0]

        # Save strategy :
        # - HuggingFace model & tokenizer are already saved in the fit() function
        # - We don't want them in the .pkl as they are heavy & already saved
        # - Also get rid of the pipe (takes too much disk space for nothing),
        #   will be reloaded automatically at first call to predict functions
        hf_model = self.model
        hf_tokenizer = self.tokenizer
        pipe = self.pipe
        self.model = None
        self.tokenizer = None
        self.pipe = None
        super().save(json_data=json_data)
        self.model = hf_model
        self.tokenizer = hf_tokenizer
        self.pipe = pipe

    def _hook_post_load_model_pkl(self):
        '''Manages a model specificities post load from a pickle file (i.e. not from standalone files)

        Raises:
            FileNotFoundError: If the HF model directory does not exist
            FileNotFoundError: If the HF tokenizer directory does not exist
        '''
        # Paths
        hf_model_dir = os.path.join(self.model_dir, 'hf_model')
        hf_tokenizer_dir = os.path.join(self.model_dir, 'hf_tokenizer')

        # Manage errors
        if not os.path.isdir(hf_model_dir):
            raise FileNotFoundError(f"Can't find HF model directory ({hf_model_dir})")
        if not os.path.isdir(hf_tokenizer_dir):
            raise FileNotFoundError(f"Can't find HF tokenizer directory ({hf_tokenizer_dir})")

        # Loading the model
        self.model = self._get_model(hf_model_dir)
        # Loading the tokenizer
        self.tokenizer = self._get_tokenizer(hf_tokenizer_dir)

    @classmethod
    def _init_new_instance_from_configs(cls, configs):
        '''Inits a new instance from a set of configurations

        Args:
            configs: a set of configurations of a model to be reloaded
        Returns:
            ModelClass: the newly generated class
        '''
        # Call parent
        model = super()._init_new_instance_from_configs(configs)

        # Try to read the following attributes from configs and, if absent, keep the current one
        for attribute in ['batch_size', 'epochs', 'validation_split', 'patience',
                          'transformer_name', 'transformer_params', 'trainer_params', 'model_max_length']:
            setattr(model, attribute, configs.get(attribute, getattr(model, attribute)))

        # Return the new model
        return model

    def _load_standalone_files(self, default_model_dir: Union[str, None] = None, hf_model_dir_path: Union[str, None] = None,
                               hf_tokenizer_dir_path: Union[str, None] = None, *args, **kwargs):
        '''Loads standalone files for a newly created model via _init_new_instance_from_configs

        Kwargs:
            default_model_dir (str): a path to look for default file paths
                                     If None, standalone files path should all be provided
            hf_model_dir_path (str): path to HF model directory.
                                If None, we'll use the default path if default_model_dir is not None
            hf_tokenizer_dir_path (str): path to HF tokenizer directory.
                                    If None, we'll use the default path if default_model_dir is not None
        Raises:
            ValueError: If at least one path is not specified and can't be inferred
            FileNotFoundError: If the HF model directory does not exist
            FileNotFoundError: If the HF tokenizer directory does not exist
        '''
        # Check if we are able to get all needed paths
        if default_model_dir is None and None in [hf_model_dir_path, hf_tokenizer_dir_path]:
            raise ValueError("At least one path is not specified and can't be inferred")

        # Retrieve file paths
        if hf_model_dir_path is None:
            hf_model_dir_path = os.path.join(default_model_dir, "hf_model")
        if hf_tokenizer_dir_path is None:
            hf_tokenizer_dir_path = os.path.join(default_model_dir, "hf_tokenizer")

        # Check paths exists
        if not os.path.isdir(hf_model_dir_path):
            raise FileNotFoundError(f"Can't find HF model directory ({hf_model_dir_path})")
        if not os.path.isdir(hf_tokenizer_dir_path):
            raise FileNotFoundError(f"Can't find HF tokenizer directory ({hf_tokenizer_dir_path})")

        # Reload model & tokenizer
        self.model = self._get_model(hf_model_dir_path)
        self.tokenizer = self._get_tokenizer(hf_tokenizer_dir_path)

        # Save hf folders in new folder (as this is skipped in save function)
        new_hf_model_dir_path = os.path.join(self.model_dir, 'hf_model')
        new_hf_tokenizer_dir_path = os.path.join(self.model_dir, 'hf_tokenizer')
        shutil.copytree(hf_model_dir_path, new_hf_model_dir_path)
        shutil.copytree(hf_tokenizer_dir_path, new_hf_tokenizer_dir_path)

    def _is_gpu_activated(self) -> bool:
        '''Checks if a GPU is used

        Returns:
            bool: whether GPU is available or not
        '''
        # Check for available GPU devices
        return torch.cuda.is_available()


# From https://discuss.huggingface.co/t/metrics-for-training-set-in-trainer/2461/5
class MetricsTrainCallback(TrainerCallback):

    def __init__(self, trainer) -> None:
        super().__init__()
        self._trainer = trainer

    # Add metrics on train dataset
    def on_epoch_end(self, args, state, control, **kwargs):
        if control.should_evaluate:
            control_copy = deepcopy(control)
            self._trainer.evaluate(eval_dataset=self._trainer.train_dataset, metric_key_prefix='train_metrics')
            return control_copy


if __name__ == '__main__':
    logger = logging.getLogger(__name__)
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")



#!/usr/bin/env python3
# Copyright (C) <2018-2021>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Libs unittest
import unittest
from unittest.mock import Mock
from unittest.mock import patch

# Utils libs
import os
import json
import numpy as np
import pandas as pd
from {{package_name}} import utils
from {{package_name}}.preprocessing import outlier_detection

# Disable logging
import logging
logging.disable(logging.CRITICAL)


class OutlierDetectionTests(unittest.TestCase):
    '''Main class to test all functions in {{package_name}}.preprocessing.outlier_detection'''


    def setUp(self):
        '''SetUp fonction'''
        # Change directory to script directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)


    def test01_check_for_outliers(self):
        '''Test of the method outlier_detection.check_for_outliers'''

        # Valids to test
        df = pd.DataFrame({
            'x': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
            'y': [1, 2, 3, 4, 5, 6, 7, 100000, 100000000, 100000000],
            'z': [-10000, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        })
        df_copy = df.copy(deep=True)
        arr = df.to_numpy()
        arr_copy = arr.copy()
        expected_outliers = np.array([-1,  1,  1,  1,  1,  1,  1, -1, -1, -1])

        # Nominal case (we just set n_neighbors to a value < nb elements)
        outliers = outlier_detection.check_for_outliers(df, n_neighbors=3)
        outliers_bis = outlier_detection.check_for_outliers(arr, n_neighbors=3)
        # We can't test returned values as IsolationForest has a random part

        # Only IsolationForest
        outliers = outlier_detection.check_for_outliers(df, n_neighbors=0)
        outliers_bis = outlier_detection.check_for_outliers(arr, n_neighbors=0)
        # We can't test returned values as IsolationForest has a random part

        # Only LocalOutlierFactor
        outliers = outlier_detection.check_for_outliers(df, n_estimators=0, n_neighbors=3)
        outliers_bis = outlier_detection.check_for_outliers(arr, n_estimators=0, n_neighbors=3)
        # We can test returned values as LocalOutlierFactor does not have a random part
        np.testing.assert_array_equal(outliers, expected_outliers)
        np.testing.assert_array_equal(outliers_bis, expected_outliers)
        pd.testing.assert_frame_equal(df, df_copy)
        np.testing.assert_array_equal(arr, arr_copy)


# Perform tests
if __name__ == '__main__':
    # Start tests
    unittest.main()



#!/usr/bin/env python3

## Data preprocessing
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Ex: python 1_preprocess_data.py -f dataset_train.csv --target_cols Survived


import os
import gc
import time
import logging
import argparse
import pandas as pd
import dill as pickle
from pathlib import Path
from datetime import datetime
from typing import Union, List, Tuple

from {{package_name}} import utils
from {{package_name}}.preprocessing import preprocess

# Get logger
logger = logging.getLogger('{{package_name}}.1_preprocess_data')


def main(filenames: List[str], preprocessing: Union[str, None], target_cols: List[Union[str, int]],
         sep: str = '{{default_sep}}', encoding: str = '{{default_encoding}}', overwrite: bool = False) -> None:
    '''Preprocesses some datasets

    Idea:
        - For each file/dataset:
            - We retrieve a NEW pipeline
            - We `fit_transform` it on the dataset
            - We save the pipeline
            - We save the preprocessed dataset

    /!\\ `validation` dataset MUST NOT be preprocessed here, as this is the fitting part /!\\
    To apply an existing pipeline to another file, you should use 2_apply_existing_pipeline.py.

    Args:
        filenames (list<str>): Datasets filenames (actually paths relative to {{package_name}}-data)
        preprocessing (str): Preprocessing to be applied. All preprocessings are applied if None.
        target_cols (list<str|int>): List of target columns (i.e. Y).
            Warning: If you have several targets, your preprocessing must be compatible.
    Kwargs:
        sep (str): Separator to use with the .csv files
        encoding (str): Encoding to use with the .csv files
        overwrite (bool): Whether to allow overwriting
    Raises:
        ValueError: if the preprocessing is not known
        FileNotFoundError: If a given file does not exist in {{package_name}}-data
    '''
    logger.info("Data preprocessing ...")

    ##############################################
    # Manage preprocessing pipelines
    ##############################################

    # Get preprocess dictionnary
    pipelines_dict = preprocess.get_pipelines_dict()

    # Get preprocessing(s) to be applied
    if preprocessing is not None:
        # Check presence in pipelines_dict
        if preprocessing not in pipelines_dict.keys():
            raise ValueError(f"The given preprocessing {preprocessing} is not known.")
        preprocessing_list = [preprocessing]
    # By default, we apply every preprocessings
    else:
        preprocessing_list = list(pipelines_dict.keys())


    ##############################################
    # Process each file, one by one
    ##############################################

    # Process each file, one by one
    for filename in filenames:

        # Get paths
        data_path = utils.get_data_path()
        dataset_path = os.path.join(data_path, filename)
        if not os.path.exists(dataset_path):
            raise FileNotFoundError(f"The file {dataset_path} does not exist")

        # For each preprocess: get the dataframe, `fit_transform` & save both file and fitted pipeline
        for preprocess_str in preprocessing_list:

            # 'no_preprocess' must be ignored
            if preprocess_str == 'no_preprocess':
                continue

            gc.collect()  # Fix some OOM in case of huge datasets being preprocessed
            
            # Output path
            dataset_preprocessed_path = os.path.join(data_path, f'{Path(filename).stem}_{preprocess_str}.csv')

            if os.path.exists(dataset_preprocessed_path) and not overwrite:
                logger.info(f'{dataset_preprocessed_path} already exists. Pass.')
                continue
            else:
                logger.info(f'Applying preprocessing {preprocess_str} on filename {filename}')

            # Get pipeline -> we create a new pipeline for each file
            preprocess_pipeline = preprocess.get_pipeline(preprocess_str)
            # Get dataset
            df = pd.read_csv(dataset_path, sep=sep, encoding=encoding)
            df_sample = df.sample(min(100, df.shape[0]))  # Keep samples to be saved alongside the pipeline (used by explainers)
            # Split X, y
            y = df[target_cols]
            X = df.drop(target_cols, axis=1)
            # Apply pipeline
            new_X = preprocess_pipeline.fit_transform(X, y)
            # Try to retrieve new columns name (experimental)
            new_df = pd.DataFrame(new_X)
            new_df = preprocess.retrieve_columns_from_pipeline(new_df, preprocess_pipeline)
            # Reinject y -> check if a new column does not have the same name as one of the target column
            for col in target_cols:
                if col in new_df.columns:
                    new_df.rename(columns={col: f'new_{col}'}, inplace=True)
            new_df[target_cols] = y

            # Save pipeline
            # Idea: pipelines are saved in {{package_name}}-pipelines and should be reloaded to preprocess `valid` dataset.
            # Moreover, they will be saved alongside models during the training part. Hence, the models will be independent.
            pipeline_dir, pipeline_name = get_pipeline_dir(preprocess_str)
            pipeline_path = os.path.join(pipeline_dir, 'pipeline.pkl')
            # We save the pipeline as a dictionnary (pipeline object + preprocess name)
            pipeline_dict = {'preprocess_pipeline': preprocess_pipeline, 'preprocess_str': preprocess_str}
            with open(pipeline_path, 'wb') as f:
                pickle.dump(pipeline_dict, f)
            # We also save a readable file alongside the pkl file (only informative)
            info_path = os.path.join(pipeline_dir, 'pipeline.info')
            with open(info_path, 'w', encoding='{{default_encoding}}') as f:
                f.write(f"'preprocess_str': {preprocess_str}\n")
                f.write(f"'preprocess_pipeline': {str(preprocess_pipeline)}")
            # Save original dataset sample
            sample_path = os.path.join(pipeline_dir, 'dataset_sample.csv')
            utils.to_csv(df_sample, sample_path, sep=sep, encoding=encoding)

            # Save preprocessed dataframe ({{default_encoding}}, '{{default_sep}}')
            # First line is a metadata with the name of the pipeline file
            utils.to_csv(new_df, dataset_preprocessed_path, first_line=f'#{pipeline_name}', sep=sep, encoding=encoding)


def get_pipeline_dir(preprocess_str: str) -> Tuple[str, str]:
    '''Retrieves a new directory to save a pipeline

    Args:
        preprocess_str (str): name of the preprocessing being used
    Returns:
        str: absolute path to the directory
        str: name of the pipeline (= name of the directory)
    '''
    pipelines_path = utils.get_pipelines_path()
    pipeline_name = datetime.now().strftime(f"{preprocess_str}_%Y_%m_%d-%H_%M_%S")
    pipeline_dir = os.path.join(pipelines_path, pipeline_name)
    if os.path.isdir(pipeline_dir):
        # Trick : if the directory already exists (two preprocess in the same second), we wait 1 second so that the 'date' changes...
        time.sleep(1)
        return get_pipeline_dir(preprocess_str)
    else:
        os.makedirs(pipeline_dir)
    return pipeline_dir, pipeline_name


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--filenames', nargs='+', required=True, help='Datasets filenames (actually paths relative to {{package_name}}-data).')
    parser.add_argument('-p', '--preprocessing', default=None, help='Preprocessing to be applied. All preprocessings are applied if None.')
    parser.add_argument('--target_cols', nargs='+', required=True, help='Y columns.')
    parser.add_argument('--sep', default='{{default_sep}}', help="Separator to use with the .csv files.")
    parser.add_argument('--encoding', default="{{default_encoding}}", help="Encoding to use with the .csv files.")
    parser.add_argument('--overwrite', action='store_true', help="Whether to allow overwriting")
    args = parser.parse_args()
    main(filenames=args.filenames, preprocessing=args.preprocessing, target_cols=args.target_cols, sep=args.sep, encoding=args.encoding, overwrite=args.overwrite)
